# ruff: noqa: ERA001

import dataclasses
import logging
import typing
from collections import defaultdict
from dataclasses import dataclass, field
from inspect import getframeinfo, stack
from pathlib import Path
from typing import Any, Optional, Tuple, cast

from gwproto import Message
from gwproto.messages import CommEvent, EventBase, EventT, PingMessage
from paho.mqtt.client import MQTT_ERR_CONN_LOST, MQTT_ERR_SUCCESS, MQTTMessageInfo
from result import Ok, Result

from gwproactor import Proactor
from gwproactor.config import LoggerLevels
from gwproactor.config.proactor_config import ProactorConfig
from gwproactor.links import LinkManager, LinkState, MQTTClients, MQTTClientWrapper
from gwproactor.message import (
    DBGCommands,
    DBGPayload,
    MQTTReceiptPayload,
    MQTTSubackPayload,
)
from gwproactor.stats import LinkStats, ProactorStats


def split_subscriptions(client_wrapper: MQTTClientWrapper) -> Tuple[int, Optional[int]]:
    for topic, qos in client_wrapper.subscription_items():
        MQTTClientWrapper.subscribe(client_wrapper, topic, qos)
    return MQTT_ERR_SUCCESS, None


def caller_str(depth: int = 3) -> str:
    caller = getframeinfo(stack()[depth][0])
    return f"{Path(caller.filename).relative_to(Path.cwd())}:{caller.lineno}, {caller.function}()"


def assert_count(
    exp: Optional[int | tuple[int | None, int | None] | typing.Sequence[int | None]],
    got: int,
    tag: str = "",
    err_str: str = "",
) -> None:
    err_str = f"{tag}  exp: {exp}  got: {got}  {caller_str()}\n{err_str}"
    if exp is not None:
        if isinstance(exp, int):
            assert got == exp, err_str
        else:
            if isinstance(exp[0], int):
                assert got >= exp[0], err_str
            if isinstance(exp[1], int):
                assert got <= exp[1], err_str


@dataclass
class RecorderLinkStats(LinkStats):
    comm_events: list[CommEvent] = field(default_factory=list)
    forwarded: dict[str, int] = field(default_factory=lambda: defaultdict(int))
    event_counts: dict[str, dict[str, int]] = field(
        default_factory=lambda: defaultdict(lambda: defaultdict(int))
    )

    def __str__(self) -> str:
        s = super().__str__()
        if self.comm_events:
            s += "\n  Comm events:"
            for comm_event in self.comm_events:
                copy_event = comm_event.model_copy(
                    update={"MessageId": comm_event.MessageId[:6] + "..."}
                )
                s += f"\n    {str(copy_event)[:154]}"
        if self.forwarded:
            s += "\n  Forwarded events *sent* by type:"
            for message_type in sorted(self.forwarded):
                s += f"\n    {self.forwarded[message_type]:3d}: [{message_type}]"
        if self.event_counts:
            s += "\n  Events *received* by src and type:"
            for event_src in sorted(self.event_counts):
                s += f"\n    src: {event_src}"
                forwards_from_src = self.event_counts[event_src]
                for message_type in sorted(forwards_from_src):
                    s += f"\n      {forwards_from_src[message_type]:3d}: [{message_type}]"
        return s


class RecorderStats(ProactorStats):
    @classmethod
    def make_link(cls, link_name: str) -> RecorderLinkStats:
        return RecorderLinkStats(link_name)

    def link(self, name: str) -> RecorderLinkStats:
        return typing.cast(RecorderLinkStats, super().link(name))


@dataclass
class _PausedAck:
    link_name: str
    message: Message[Any]
    qos: int
    context: Optional[Any]


class RecorderLinks(LinkManager):
    acks_paused: bool
    needs_ack: list[_PausedAck]

    # noinspection PyMissingConstructor
    def __init__(self, other: LinkManager) -> None:
        self.__dict__ = other.__dict__
        self.acks_paused = False
        self.needs_ack = []

    def link(self, name: str) -> LinkState:
        link = self._states.link(name)
        if link is None:
            raise RuntimeError(f"Link {name} not found.")
        return link

    @property
    def in_flight_events(self) -> dict[str, EventBase]:
        return self._in_flight_events

    @property
    def num_in_flight(self) -> int:
        return len(self._in_flight_events)

    def publish_message(
        self,
        link_name: str,
        message: Message[Any],
        qos: int = 0,
        context: Any = None,
        *,
        topic: str = "",
        use_link_topic: bool = False,
    ) -> MQTTMessageInfo:
        if self.acks_paused:
            self.needs_ack.append(_PausedAck(link_name, message, qos, context))
            return MQTTMessageInfo(-1)
        return super().publish_message(
            link_name,
            message,
            qos=qos,
            context=context,
            topic=topic,
            use_link_topic=use_link_topic,
        )

    def release_acks(self, clear: bool = False, num_to_release: int = -1) -> int:
        # self._logger.info(
        #     f"++release_acks: clear:{clear}  num_to_release:{num_to_release}"
        # )
        # path_dbg = 0
        if clear or num_to_release < 1:
            # path_dbg |= 0x00000001
            self.acks_paused = False
            needs_ack = self.needs_ack
            self.needs_ack = []
        else:
            # path_dbg |= 0x00000002
            num_to_release = min(num_to_release, len(self.needs_ack))
            needs_ack = self.needs_ack[:num_to_release]
            self.needs_ack = self.needs_ack[num_to_release:]
            # self._logger.info(f"needs_ack: {needs_ack}")
            # self._logger.info(f"self.needs_ack: {self.needs_ack}")
        if not clear:
            # path_dbg |= 0x00000004
            for paused_ack in needs_ack:
                # path_dbg |= 0x00000008
                super().publish_message(**dataclasses.asdict(paused_ack))  # noqa
        # self._logger.info(
        #     f"--release_acks: clear:{clear}  num_to_release:{num_to_release}  path:0x{path_dbg:08X}"
        # )
        return len(needs_ack)

    def generate_event(self, event: EventT) -> Result[bool, Exception]:
        if not event.Src:
            event.Src = self.publication_name
        if isinstance(event, CommEvent) and event.Src == self.publication_name:
            cast(
                RecorderLinkStats, self._stats.link(event.PeerName)
            ).comm_events.append(event)
        if event.Src != self.publication_name and event.Src in self._stats.links:
            cast(RecorderLinkStats, self._stats.link(event.Src)).forwarded[
                event.TypeName
            ] += 1
        return super().generate_event(event)


class InstrumentedProactor(Proactor):
    _subacks_paused: dict[str, bool]
    _subacks_available: dict[str, list[Message[Any]]]
    _mqtt_messages_dropped: dict[str, bool]
    DELIMIT_CHAR = "#"
    DELIMIT_STR = DELIMIT_CHAR * 150

    def __init__(self, config: ProactorConfig) -> None:
        super().__init__(config)
        self._subacks_paused = defaultdict(bool)
        self._subacks_available = defaultdict(list)
        self._mqtt_messages_dropped = defaultdict(bool)
        self._links = RecorderLinks(self.links)

    @classmethod
    def make_stats(cls) -> RecorderStats:
        return RecorderStats()

    @property
    def links(self) -> RecorderLinks:
        return typing.cast(RecorderLinks, self._links)

    @property
    def stats(self) -> RecorderStats:
        return typing.cast(RecorderStats, self._stats)

    @property
    def recorder_links(self) -> RecorderLinks:
        return typing.cast(RecorderLinks, self._links)

    @property
    def needs_ack(self) -> list[_PausedAck]:
        return self.recorder_links.needs_ack

    def force_mqtt_disconnect(self, client_name: str) -> None:
        mqtt_client = self.mqtt_client_wrapper(client_name).mqtt_client
        # noinspection PyProtectedMember
        mqtt_client._loop_rc_handle(MQTT_ERR_CONN_LOST)  # noqa

    def _process_mqtt_message(
        self, mqtt_receipt_message: Message[MQTTReceiptPayload]
    ) -> Result[Message[Any], Exception]:
        if self._mqtt_messages_dropped[mqtt_receipt_message.Payload.client_name]:
            return Ok(mqtt_receipt_message)
        match decoded_result := super()._process_mqtt_message(mqtt_receipt_message):
            case Ok(decoded):
                match decoded.Payload:
                    case EventBase() as event:
                        stats = cast(
                            RecorderLinkStats,
                            self._stats.link(mqtt_receipt_message.Payload.client_name),
                        )
                        stats.event_counts[event.Src][event.TypeName] += 1
        return decoded_result

    def subacks_paused(self, client_name: str) -> bool:
        return self._subacks_paused[client_name]

    def num_subacks_available(self, client_name: str) -> int:
        return len(self._subacks_available[client_name])

    def clear_subacks(self, client_name: str) -> None:
        self._subacks_available[client_name] = []

    def mqtt_messages_dropped(self, client_name: str) -> bool:
        return self._mqtt_messages_dropped[client_name]

    def upstream_subacks_paused(self) -> bool:
        return self.subacks_paused(self.upstream_client)

    def num_upstream_subacks_available(self) -> int:
        return self.num_subacks_available(self.upstream_client)

    def clear_upstream_subacks(self) -> None:
        self._subacks_available[self.upstream_client] = []

    def upstream_mqtt_messages_dropped(self) -> bool:
        return self.mqtt_messages_dropped(self.upstream_client)

    def split_client_subacks(self, client_name: str) -> None:
        client_wrapper = self.mqtt_client_wrapper(client_name)

        def member_split_subscriptions() -> Tuple[int, Optional[int]]:
            return split_subscriptions(client_wrapper)

        client_wrapper.subscribe_all = member_split_subscriptions  # type: ignore[method-assign]

    def restore_client_subacks(self, client_name: str) -> None:
        client_wrapper = self.mqtt_client_wrapper(client_name)
        client_wrapper.subscribe_all = MQTTClientWrapper.subscribe_all  # type: ignore[method-assign, assignment]

    def pause_subacks(self, client_name: str) -> None:
        self._subacks_paused[client_name] = True

    def pause_upstream_subacks(self) -> None:
        self.pause_subacks(self.upstream_client)

    def release_subacks(self, client_name: str, num_released: int = -1) -> None:
        if self._receive_queue is None:
            raise RuntimeError(
                "ERROR. release_subacks() called before Proactor started."
            )
        self._subacks_paused[client_name] = False
        if num_released < 0:
            num_released = len(self._subacks_available[client_name])
        release = self._subacks_available[client_name][:num_released]
        remaining = self._subacks_available[client_name][num_released:]
        self._subacks_available[client_name] = remaining
        for message in release:
            self._receive_queue.put_nowait(message)

    def release_upstream_subacks(self, num_released: int = -1) -> None:
        self.release_subacks(self.upstream_client, num_released)

    async def async_process_message(self, message: Message[Any]) -> None:
        if (
            isinstance(message.Payload, MQTTSubackPayload)
            and self._subacks_paused[message.Payload.client_name]
        ):
            self._subacks_available[message.Payload.client_name].append(message)
        else:
            await super().async_process_message(message)

    def pause_acks(self) -> None:
        self.recorder_links.acks_paused = True

    def release_acks(self, clear: bool = False, num_to_release: int = -1) -> int:
        return typing.cast(RecorderLinks, self._links).release_acks(
            clear, num_to_release=num_to_release
        )

    def set_ack_timeout_seconds(self, delay: float) -> None:
        self.links.ack_manager._default_delay_seconds = delay  # noqa: SLF001

    def restore_ack_timeout_seconds(self) -> None:
        self.links.ack_manager._default_delay_seconds = (  # noqa: SLF001
            self.settings.proactor.ack_timeout_seconds
        )

    def drop_mqtt(self, client_name: str, drop: bool) -> None:
        self._mqtt_messages_dropped[client_name] = drop

    def summary_str(self) -> str:
        s = str(self.stats)
        s += "\nEvents:\n"
        s += f"  pending: {self.links.num_pending}\n"
        s += f"  in-flight: {self.links.num_in_flight}\n"
        s += f"  persisted: {self.event_persister.num_persists}\n"
        s += f"  retrieved: {self.event_persister.num_retrieves}\n"
        s += f"  cleared: {self.event_persister.num_clears}\n"
        s += "Link states:\n"
        for link_name in self.stats.links:
            link_state = self._links.link_state(link_name)
            if link_state is None:
                raise KeyError("ERROR. LinkManager has no link state for <{link_name}>")
            s += f"  {link_name:10s}  {link_state.value}\n"
        s += self.links.subscription_str().lstrip()
        s += "Pending acks:\n"
        for link_name in self.stats.links:
            s += f"  {link_name:10s}  {self._links.num_acks(link_name):3d}\n"
        s += self._links.get_reuploads_str() + "\n"
        s += f"Paused acks: {len(self.needs_ack)}\n"
        s += "Paused Subacks:"
        for link_name in self.stats.links:
            s += (
                f"  {link_name:10s}  "
                f"subacks paused: {self._subacks_paused[link_name]}  "
                f"subacks available: {len(self._subacks_available[link_name])}\n"
            )
        return s

    def summarize(self) -> None:
        self._logger.info(self.summary_str())

    def delimit(self, text: str = "", log_level: int = logging.INFO) -> None:
        if self._logger.isEnabledFor(log_level):
            self._logger.log(
                log_level,
                f"\n\n{self.DELIMIT_STR}\n"
                f"{self.DELIMIT_CHAR}  {text}\n"
                f"{self.DELIMIT_STR}\n",
            )

    def force_ping(self, client_name: str) -> None:
        self._links.publish_message(client_name, PingMessage(Src=self.publication_name))

    @property
    def mqtt_clients(self) -> MQTTClients:
        return self._links.mqtt_clients()

    def mqtt_client_wrapper(self, client_name: str) -> MQTTClientWrapper:
        return self._links.mqtt_client_wrapper(client_name)

    def mqtt_subscriptions(self, client_name: str) -> list[str]:
        return [
            item[0]
            for item in self.mqtt_client_wrapper(client_name).subscription_items()
        ]

    def all_mqtt_subscriptions(self) -> list[str]:
        subscriptions = []
        for client_name in self.mqtt_clients.clients:
            subscriptions.extend(self.mqtt_subscriptions(client_name))
        return subscriptions

    def send_dbg(
        self,
        client_name: str,
        message_summary: int = -1,
        lifecycle: int = -1,
        comm_event: int = -1,
        command: Optional[DBGCommands | str] = None,
    ) -> None:
        if isinstance(command, str):
            command = DBGCommands(command)
        self.send_threadsafe(
            Message(
                Src=self.name,
                Dst=client_name,
                Payload=DBGPayload(
                    Levels=LoggerLevels(
                        message_summary=message_summary,
                        lifecycle=lifecycle,
                        comm_event=comm_event,
                    ),
                    Command=command,
                ),
            )
        )

    def mqtt_quiescent(self) -> bool:
        mqtt_quiescent = getattr(super(), "mqtt_quiescent()", None)
        if mqtt_quiescent is not None:
            return typing.cast(bool, mqtt_quiescent())
        link = self._links.link(self.upstream_client)
        return link is None or link.active_for_send()

    def _call_super_if_present(self, function_name: str) -> None:
        if hasattr(super(), function_name):
            getattr(super(), function_name)()

    def disable_derived_events(self) -> None:
        self._call_super_if_present("disable_dervived_events")

    def enable_derived_events(self) -> None:
        self._call_super_if_present("enable_dervived_events")

    def events_at_rest(
        self,
        num_pending: int = 0,
        *,
        exact_pending: bool = True,
        num_persists: Optional[int] = None,
        exact_persists: bool = True,
    ) -> bool:
        if exact_pending:
            pending_check = self.links.num_pending == num_pending
        else:
            pending_check = self.links.num_pending >= num_pending
        if num_persists is None:
            persist_check = True
        elif exact_persists:
            persist_check = self.event_persister.num_persists == num_persists
        else:
            persist_check = self.event_persister.num_persists >= num_persists
        return pending_check and self.links.num_in_flight == 0 and persist_check

    def assert_event_counts(
        self,
        *,
        num_pending: Optional[int | tuple[int | None, int | None]] = 0,
        num_in_flight: Optional[int | tuple[int | None, int | None]] = 0,
        all_pending: bool = False,
        num_persists: Optional[int | tuple[int | None, int | None]] = None,
        num_retrieves: Optional[int | tuple[int | None, int | None]] = None,
        num_clears: Optional[int | tuple[int | None, int | None]] = None,
        all_clear: bool = False,
        tag: str = "",
        err_str: str = "",
    ) -> None:
        assert_count(num_pending, self.links.num_pending, tag + " num_pending", err_str)
        assert_count(
            num_in_flight, self.links.num_in_flight, tag + " num_in_flight", err_str
        )
        p = self.event_persister
        if all_pending:
            assert_count(num_pending, p.num_persists, tag + " num_persists", err_str)
            assert_count(0, p.num_retrieves, tag + " num_retrieves", err_str)
            assert_count(0, p.num_clears, tag + " num_clears", err_str)
        if all_clear:
            if num_pending != 0:
                raise ValueError(
                    f"ERROR. all_clear is True but num_pending ({num_pending}) != 0"
                )
            if num_retrieves is None:
                num_retrieves = num_persists
            if num_clears is None:
                num_clears = num_persists
        assert_count(num_persists, p.num_persists, tag + " num_persists", err_str)
        assert_count(num_retrieves, p.num_retrieves, tag + " num_retrieves", err_str)
        assert_count(num_clears, p.num_clears, tag + " num_clears", err_str)



"""Actor: A partial implementation of ActorInterface which supplies the trivial implementations.

SyncThreadActor: An actor which orchestrates starting, stopping and communicating with a passed in
SyncAsyncInteractionThread
"""

from abc import ABC
from typing import Any, Generic, Sequence, TypeVar

from gwproto import Message, ShNode
from result import Ok, Result

from gwproactor.callbacks import ProactorCallbackInterface
from gwproactor.codecs import CodecFactory
from gwproactor.proactor_interface import (
    ActorInterface,
    AppInterface,
    Communicator,
    MonitoredName,
)
from gwproactor.sync_thread import SyncAsyncInteractionThread


class Actor(ActorInterface, Communicator, ABC):
    _node: ShNode

    def __init__(self, name: str, services: AppInterface) -> None:
        self._node = services.hardware_layout.node(name)
        super().__init__(name, services)

    @classmethod
    def instantiate(
        cls, name: str, services: "AppInterface", **constructor_args: Any
    ) -> "ActorInterface":
        return cls(
            name,
            services,
            **constructor_args,  # noqa
        )

    @property
    def name(self) -> str:
        return self._name

    @property
    def node(self) -> ShNode:
        return self._node

    def init(self) -> None:
        """Called after constructor so derived functions can be used in setup."""


SyncThreadT = TypeVar("SyncThreadT", bound=SyncAsyncInteractionThread)


class SyncThreadActor(Actor, Generic[SyncThreadT]):
    _sync_thread: SyncAsyncInteractionThread

    def __init__(
        self,
        name: str,
        services: AppInterface,
        sync_thread: SyncAsyncInteractionThread,
    ) -> None:
        super().__init__(name, services)
        self._sync_thread = sync_thread

    def process_message(self, message: Message[Any]) -> Result[bool, Exception]:
        raise ValueError(
            f"Error. {self.__class__.__name__} does not process any messages. Received {message.Header}"
        )

    def send_driver_message(self, message: Any) -> None:
        self._sync_thread.put_to_sync_queue(message)

    def start(self) -> None:
        if (
            self.services.event_loop is None
            or self.services.async_receive_queue is None
        ):
            raise ValueError("ERROR. Actor started before ServicesInterface started")
        self._sync_thread.set_async_loop_and_start(
            self.services.event_loop, self.services.async_receive_queue
        )

    def stop(self) -> None:
        self._sync_thread.request_stop()

    async def join(self) -> None:
        await self._sync_thread.async_join()

    @property
    def monitored_names(self) -> Sequence[MonitoredName]:
        monitored_names = []
        if self._sync_thread.pat_timeout is not None:
            monitored_names.append(
                MonitoredName(self.name, self._sync_thread.pat_timeout)
            )
        return monitored_names


class PrimeActor(ProactorCallbackInterface, Actor):
    def __init__(self, name: str, services: AppInterface) -> None:
        super().__init__(name, services)
        services.add_callbacks(self)
        services.add_communicator(self)

    def process_message(self, _: Message[Any]) -> Result[bool, Exception]:
        return Ok(value=True)

    def start(self) -> None:
        pass

    def stop(self) -> None:
        pass

    async def join(self) -> None:
        pass

    @classmethod
    def get_codec_factory(cls) -> CodecFactory:
        return CodecFactory()


class NullPrimeActor(PrimeActor): ...



"""MQTT infrastructure providing support for multiple MTQTT clients

Main current limitation: each interaction between asyncio code and the mqtt clients must either have thread locking
(as is provided inside paho for certain functions such as publish()) or an explicit message based API.

"""

import asyncio
import contextlib
import enum
import logging
import ssl
import threading
import typing
import uuid
from typing import Any, Dict, List, NamedTuple, Optional, Set, Tuple, Union, cast

from paho.mqtt.client import (
    MQTT_ERR_SUCCESS,
    ConnectFlags,
    DisconnectFlags,
    MQTTMessage,
    MQTTMessageInfo,
)
from paho.mqtt.client import Client as PahoMQTTClient
from paho.mqtt.enums import CallbackAPIVersion, MQTTErrorCode
from paho.mqtt.properties import Properties as MQTTProperties
from paho.mqtt.reasoncodes import ReasonCode

from gwproactor import config
from gwproactor.links.link_settings import LinkConfig
from gwproactor.message import (
    MQTTConnectFailMessage,
    MQTTConnectMessage,
    MQTTDisconnectMessage,
    MQTTProblemsMessage,
    MQTTReceiptMessage,
    MQTTSubackMessage,
    MQTTSubackPayload,
)
from gwproactor.problems import Problems
from gwproactor.sync_thread import AsyncQueueWriter, responsive_sleep


class QOS(enum.IntEnum):
    AtMostOnce = 0
    AtLeastOnce = 1
    ExactlyOnce = 2


class Subscription(NamedTuple):
    Topic: str
    Qos: QOS


class MQTTClientWrapper:
    _client_name: str
    topic_dst: str
    _client_config: config.MQTTClient
    _client: PahoMQTTClient
    _stop_requested: bool
    _receive_queue: AsyncQueueWriter
    _subscriptions: Dict[str, int]
    _pending_subscriptions: Set[str]
    _pending_subacks: Dict[int, List[str]]

    def __init__(
        self,
        client_name: str,
        topic_dst: str,
        client_config: config.MQTTClient,
        receive_queue: AsyncQueueWriter,
    ) -> None:
        self._client_name = client_name
        self.topic_dst = topic_dst
        self._client_config = client_config
        self._receive_queue = receive_queue
        self._client = PahoMQTTClient(
            callback_api_version=CallbackAPIVersion.VERSION2,
            client_id="-".join(str(uuid.uuid4()).split("-")[:-1]),
        )
        if self._client_config.username is not None:
            self._client.username_pw_set(
                username=self._client_config.username,
                password=self._client_config.password.get_secret_value(),
            )
        tls_config = self._client_config.tls
        if tls_config.use_tls:
            try:
                self._client.tls_set(
                    ca_certs=str(tls_config.paths.ca_cert_path),
                    certfile=str(tls_config.paths.cert_path),
                    keyfile=str(tls_config.paths.private_key_path),
                    cert_reqs=tls_config.cert_reqs,
                    tls_version=ssl.PROTOCOL_TLS_CLIENT,
                    ciphers=tls_config.ciphers,
                    keyfile_password=tls_config.keyfile_password.get_secret_value(),
                )
            except Exception as e:
                raise RuntimeError(
                    f"ERROR. Client.tls_set() caught exception <{e}> "
                    f"from tls_config: {tls_config}"
                ) from e

        self._client.on_message = self.on_message
        self._client.on_connect = self.on_connect
        self._client.on_connect_fail = self.on_connect_fail
        self._client.on_disconnect = self.on_disconnect
        self._client.on_subscribe = self.on_subscribe
        self._subscriptions = {}
        self._pending_subscriptions = set()
        self._pending_subacks = {}
        self._thread = threading.Thread(
            target=self._client_thread,
            name=f"MQTT-client-thread-{self._client_name}",
            daemon=True,
        )
        self._stop_requested = False

    def _client_thread(self) -> None:
        max_back_off = 1024
        backoff = 1
        while not self._stop_requested:
            try:
                self._client.connect(
                    self._client_config.host, port=self._client_config.effective_port()
                )
                self._client.loop_forever(retry_first_connection=True)
            except Exception as e:  # noqa: BLE001
                self._receive_queue.put(
                    MQTTProblemsMessage(
                        client_name=self._client_name, problems=Problems(errors=[e])
                    )
                )
            finally:
                with contextlib.suppress(Exception):
                    self._client.disconnect()
            if not self._stop_requested:
                if backoff >= max_back_off:
                    backoff = 1
                else:
                    backoff = min(backoff * 2, max_back_off)
                responsive_sleep(
                    self,
                    backoff,
                    running_field_name="_stop_requested",
                    running_field=False,
                )

    def start(self) -> None:
        self._thread.start()

    def stop(self) -> None:
        self._stop_requested = True
        try:
            self._client.disconnect()
            self._thread.join()
        except:  # noqa: E722, S110
            pass

    def publish(self, topic: str, payload: bytes, qos: int) -> MQTTMessageInfo:
        return self._client.publish(topic, payload, qos)

    def subscribe(self, topic: str, qos: int) -> tuple[int, Optional[int]]:
        self._subscriptions[topic] = qos
        self._pending_subscriptions.add(topic)
        subscribe_result: tuple[MQTTErrorCode, int | None] = self._client.subscribe(
            topic, qos
        )
        if subscribe_result[0] == MQTT_ERR_SUCCESS:
            self._pending_subacks[typing.cast(int, subscribe_result[1])] = [topic]
        return typing.cast(tuple[int, Optional[int]], subscribe_result)

    def subscribe_all(self) -> tuple[int, Optional[int]]:
        subscribe_result: tuple[int, Optional[int]]
        if self._subscriptions:
            topics = list(self._subscriptions.keys())
            for topic in topics:
                self._pending_subscriptions.add(topic)
            subscribe_result = self._client.subscribe(
                list(self._subscriptions.items()), 0
            )
            if (
                subscribe_result[0] == MQTT_ERR_SUCCESS
                and subscribe_result[1] is not None
            ):
                self._pending_subacks[subscribe_result[1]] = topics
        else:
            subscribe_result = MQTT_ERR_SUCCESS, None
        return subscribe_result

    def unsubscribe(self, topic: str) -> Tuple[int, Optional[int]]:
        self._subscriptions.pop(topic, None)
        return typing.cast(tuple[int, Optional[int]], self._client.unsubscribe(topic))

    def connected(self) -> bool:
        return self._client.is_connected()

    def num_subscriptions(self) -> int:
        return len(self._subscriptions)

    def num_pending_subscriptions(self) -> int:
        return len(self._pending_subscriptions)

    def subscribed(self) -> bool:
        return self.connected() and (
            not self.num_subscriptions() or not self.num_pending_subscriptions()
        )

    def subscription_items(self) -> list[Tuple[str, int]]:
        return list(cast(list[Tuple[str, int]], self._subscriptions.items()))

    @property
    def mqtt_client(self) -> PahoMQTTClient:
        return self._client

    def on_message(self, _: Any, userdata: Any, message: MQTTMessage) -> None:
        self._receive_queue.put(
            MQTTReceiptMessage(
                client_name=self._client_name,
                userdata=userdata,
                message=message,
            )
        )

    def handle_suback(self, suback: MQTTSubackPayload) -> int:
        topics = self._pending_subacks.pop(suback.mid, [])
        if topics:
            for topic in topics:
                self._pending_subscriptions.remove(topic)
        return len(self._pending_subscriptions)

    def on_subscribe(
        self,
        _: PahoMQTTClient,
        userdata: Any,
        mid: int,
        reason_codes: typing.Sequence[ReasonCode],
        _properties: MQTTProperties,
    ) -> None:
        self._receive_queue.put(
            MQTTSubackMessage(
                client_name=self._client_name,
                userdata=userdata,
                mid=mid,
                reason_codes=reason_codes,
            )
        )

    def on_connect(
        self,
        _: Any,
        userdata: Any,
        flags: ConnectFlags,
        reason_code: ReasonCode,
        _properties: Optional[MQTTProperties] = None,
    ) -> None:
        self._receive_queue.put(
            MQTTConnectMessage(
                client_name=self._client_name,
                userdata=userdata,
                flags=flags,
                rc=reason_code,
            )
        )

    def on_connect_fail(self, _: Any, userdata: Any) -> None:
        self._receive_queue.put(
            MQTTConnectFailMessage(
                client_name=self._client_name,
                userdata=userdata,
            )
        )

    def on_disconnect(
        self,
        _: PahoMQTTClient,
        userdata: Any,
        _flags: DisconnectFlags,
        rc: ReasonCode,
        _properties: Optional[MQTTProperties],
    ) -> None:
        self._pending_subscriptions = set(self._subscriptions.keys())
        self._receive_queue.put(
            MQTTDisconnectMessage(
                client_name=self._client_name,
                userdata=userdata,
                rc=rc,
            )
        )

    def enable_logger(
        self,
        logger: Optional[
            Union[logging.Logger, logging.LoggerAdapter[logging.Logger]]
        ] = None,
    ) -> None:
        self._client.enable_logger(typing.cast(logging.Logger | None, logger))

    def disable_logger(self) -> None:
        self._client.disable_logger()


class MQTTClients:
    clients: Dict[str, MQTTClientWrapper]
    _send_queue: AsyncQueueWriter
    upstream_client: str = ""
    upstream_topic_dst: str = ""
    downstream_client: str = ""

    def __init__(self) -> None:
        self._send_queue = AsyncQueueWriter()
        self.clients = {}

    def add_client(self, settings: LinkConfig) -> None:
        if settings.client_name in self.clients:
            raise ValueError(
                f"ERROR. MQTT client named {settings.client_name} already exists"
            )
        if settings.upstream:
            if self.upstream_client:
                raise ValueError(
                    f"ERROR. upstream client already set as {self.upstream_client}. "
                    f"Client {settings.client_name} may not be set as upstream."
                )
            self.upstream_client = settings.client_name
            self.upstream_topic_dst = settings.spaceheat_name
        if settings.downstream:
            if self.downstream_client:
                raise ValueError(
                    f"ERROR. primary peer client already set as {self.downstream_client}. "
                    f"Client {settings.client_name} may not be set as primary peer."
                )
            self.downstream_client = settings.client_name
        self.clients[settings.client_name] = MQTTClientWrapper(
            client_name=settings.client_name,
            topic_dst=settings.spaceheat_name,
            client_config=settings.mqtt,
            receive_queue=self._send_queue,
        )

    def publish(
        self, client: str, topic: str, payload: bytes, qos: int
    ) -> MQTTMessageInfo:
        return self.clients[client].publish(topic, payload, qos)

    def subscribe(self, client: str, topic: str, qos: int) -> Tuple[int, Optional[int]]:
        return self.clients[client].subscribe(topic, qos)

    def subscribe_all(self, client: str) -> Tuple[int, Optional[int]]:
        return self.clients[client].subscribe_all()

    def unsubscribe(self, client: str, topic: str) -> Tuple[int, Optional[int]]:
        return self.clients[client].unsubscribe(topic)

    def handle_suback(self, suback: MQTTSubackPayload) -> int:
        return self.clients[suback.client_name].handle_suback(suback)

    def stop(self) -> None:
        for client in self.clients.values():
            client.stop()

    def start(
        self, loop: asyncio.AbstractEventLoop, async_queue: asyncio.Queue[Any]
    ) -> None:
        self._send_queue.set_async_loop(loop, async_queue)
        for client in self.clients.values():
            client.start()

    def connected(self, client: str) -> bool:
        return self.clients[client].connected()

    def num_subscriptions(self, client: str) -> int:
        return self.clients[client].num_subscriptions()

    def num_pending_subscriptions(self, client: str) -> int:
        return self.clients[client].num_pending_subscriptions()

    def subscribed(self, client: str) -> bool:
        return self.clients[client].subscribed()

    def enable_loggers(
        self,
        logger: Optional[
            Union[logging.Logger, logging.LoggerAdapter[logging.Logger]]
        ] = None,
    ) -> None:
        for client_name in self.clients:
            self.clients[client_name].enable_logger(logger)

    def disable_loggers(self) -> None:
        for client_name in self.clients:
            self.clients[client_name].disable_logger()

    def client_wrapper(self, client: str) -> MQTTClientWrapper:
        return self.clients[client]

    def upstream(self) -> MQTTClientWrapper:
        return self.clients[self.upstream_client]

    def primary_peer(self) -> MQTTClientWrapper:
        return self.clients[self.downstream_client]

    def topic_dst(self, client: str) -> str:
        if client in self.clients:
            return self.clients[client].topic_dst
        return ""



import contextlib
import datetime
import re
import shutil
import subprocess
import time
from pathlib import Path
from typing import NamedTuple, Optional

from result import Err, Ok, Result

from gwproactor.persister.exceptions import (
    ContentTooLarge,
    FileExistedWarning,
    FileMissing,
    FileMissingWarning,
    PersisterError,
    ReadFailed,
    ReindexError,
    TrimFailed,
    UIDExistedWarning,
    UIDMissingWarning,
    WriteFailed,
)
from gwproactor.persister.interface import PersisterInterface
from gwproactor.problems import Problems


class _PersistedItem(NamedTuple):
    uid: str
    path: Path


class TimedRollingFilePersister(PersisterInterface):
    DEFAULT_MAX_BYTES: int = 500 * 1024 * 1024
    FILENAME_RGX: re.Pattern[str] = re.compile(r"(?P<dt>.*)\.uid\[(?P<uid>.*)].json$")
    REINDEX_PAT_SECONDS = 1.0

    _base_dir: Path
    _max_bytes: int = DEFAULT_MAX_BYTES
    _pending: dict[str, Path]
    _curr_dir: Path
    _curr_bytes: int
    _pat_watchdog_args: Optional[list[str]] = None
    _reindex_pat_seconds: float = REINDEX_PAT_SECONDS
    _num_persists: int = 0
    _num_retrieves: int = 0
    _num_clears: int = 0

    def __init__(
        self,
        base_dir: Path | str,
        max_bytes: int = DEFAULT_MAX_BYTES,
        pat_watchdog_args: Optional[list[str]] = None,
        reindex_pat_seconds: float = REINDEX_PAT_SECONDS,
    ) -> None:
        self._base_dir = Path(base_dir).resolve()
        self._max_bytes = max_bytes
        self._curr_dir = self._today_dir()
        self._curr_bytes = 0
        self._pat_watchdog_args = pat_watchdog_args
        self._reindex_pat_seconds = reindex_pat_seconds
        self._pending = {}

    @property
    def max_bytes(self) -> int:
        return self._max_bytes

    @property
    def curr_bytes(self) -> int:
        return self._curr_bytes

    @property
    def base_dir(self) -> Path:
        return self._base_dir

    @property
    def curr_dir(self) -> Path:
        return self._curr_dir

    def persist(self, uid: str, content: bytes) -> Result[bool, Problems]:
        self._num_persists += 1
        problems = Problems()
        try:
            if len(content) > self._max_bytes:
                return Err(
                    problems.add_error(
                        ContentTooLarge(
                            f"content bytes ({len(content)} > max bytes {self._max_bytes}",
                            uid=uid,
                        )
                    )
                )
            if len(content) + self._curr_bytes > self._max_bytes:
                trimmed = self._trim_old_storage(len(content))
                match trimmed:
                    case Err(trim_problems):
                        problems.add_problems(trim_problems)
                        if problems.errors:
                            return Err(problems.add_error(TrimFailed(uid=uid)))
            existing_path = self._pending.pop(uid, None)
            if existing_path is not None:
                problems.add_warning(UIDExistedWarning(uid=uid, path=existing_path))
                if existing_path.exists():
                    self._curr_bytes -= existing_path.stat().st_size
                    problems.add_warning(
                        FileExistedWarning(uid=uid, path=existing_path)
                    )
                else:
                    problems.add_warning(
                        FileMissingWarning(uid=uid, path=existing_path)
                    )
            self._roll_curr_dir()
            self._pending[uid] = self._curr_dir / self._make_name(
                datetime.datetime.now(tz=datetime.timezone.utc), uid
            )
            try:
                with self._pending[uid].open("wb") as f:
                    f.write(content)
                self._curr_bytes += len(content)
            except Exception as e:  # pragma: no cover  # noqa: BLE001
                return Err(
                    problems.add_error(e).add_error(
                        WriteFailed("Open or write failed", uid=uid, path=existing_path)
                    )
                )
        except Exception as e:  # noqa: BLE001
            return Err(
                problems.add_error(e).add_error(
                    PersisterError("Unexpected error", uid=uid)
                )
            )
        if problems:
            return Err(problems)
        return Ok()

    def _trim_old_storage(self, needed_bytes: int) -> Result[bool, Problems]:
        problems = Problems()
        last_day_dir: Optional[Path] = None
        items = list(self._pending.items())
        for uid, path in items:
            try:
                match self.clear(uid):
                    case Err(other):
                        problems.add_problems(other)
                day_dir = path.parent
                if last_day_dir is not None and last_day_dir != day_dir:
                    shutil.rmtree(last_day_dir, ignore_errors=True)
                last_day_dir = day_dir
            except Exception as e:  # noqa: BLE001
                problems.add_error(e)
                problems.add_error(
                    PersisterError("Unexpected error", uid=uid, path=path)
                )
            if self._curr_bytes <= self._max_bytes - needed_bytes:
                break
        try:
            if last_day_dir is not None and (
                not self._pending
                or next(iter(self._pending.values())).parent != last_day_dir
            ):
                shutil.rmtree(last_day_dir, ignore_errors=True)
        except Exception as e:  # pragma: no cover  # noqa: BLE001
            problems.add_error(e)
            problems.add_error(PersisterError("Unexpected error"))
        if problems:
            return Err(problems)
        return Ok()

    def clear(self, uid: str) -> Result[bool, Problems]:
        self._num_clears += 1
        problems = Problems()
        path = self._pending.pop(uid, None)
        if path:
            if path.exists():
                self._curr_bytes -= path.stat().st_size
                path.unlink()
                path_dir = path.parent
                # Remove directory if empty.
                # This is much faster than using iterdir.
                with contextlib.suppress(OSError):
                    path_dir.rmdir()
            else:
                problems.add_warning(FileMissingWarning(uid=uid, path=path))
        else:
            problems.add_warning(UIDMissingWarning(uid=uid, path=path))
        if problems:
            return Err(problems)
        return Ok()

    def pending_ids(self) -> list[str]:
        return list(self._pending.keys())

    def pending_paths(self) -> list[Path]:
        return list(self._pending.values())

    def pending_dict(self) -> dict[str, Path]:
        return dict(self._pending)

    @property
    def num_pending(self) -> int:
        return len(self._pending)

    @property
    def num_persists(self) -> int:
        return self._num_persists

    @property
    def num_retrieves(self) -> int:
        return self._num_retrieves

    @property
    def num_clears(self) -> int:
        return self._num_clears

    def __contains__(self, uid: str) -> bool:
        return uid in self._pending

    def get_path(self, uid: str) -> Optional[Path]:
        return self._pending.get(uid, None)

    def retrieve(self, uid: str) -> Result[Optional[bytes], Problems]:
        self._num_retrieves += 1
        problems = Problems()
        content: Optional[bytes] = None
        path = self._pending.get(uid, None)
        if path:
            if path.exists():
                try:
                    with path.open("rb") as f:
                        content = f.read()
                except Exception as e:  # pragma: no cover  # noqa: BLE001
                    problems.add_error(e).add_error(
                        ReadFailed("Open or read failed", uid=uid, path=path)
                    )
            else:
                problems.add_error(FileMissing(uid=uid, path=path))
        if problems:
            return Err(problems)
        return Ok(content)

    def reindex(self) -> Result[bool, Problems]:
        problems = Problems()
        self._curr_bytes = 0
        paths: list[_PersistedItem] = []
        last_pat = time.time()
        for base_dir_entry in self._base_dir.iterdir():  # noqa: PLR1702
            try:
                if base_dir_entry.is_dir() and self._is_iso_parseable(base_dir_entry):
                    for day_dir_entry in base_dir_entry.iterdir():
                        if self._pat_watchdog_args:
                            now = time.time()
                            if now > last_pat + self._reindex_pat_seconds:
                                last_pat = now
                                subprocess.run(self._pat_watchdog_args, check=True)  # noqa: S603
                        try:
                            if persisted_item := self._persisted_item_from_file_path(
                                day_dir_entry
                            ):
                                self._curr_bytes += persisted_item.path.stat().st_size
                                paths.append(persisted_item)
                        except Exception as e:  # noqa: BLE001
                            problems.add_error(e).add_error(
                                ReindexError(path=day_dir_entry)
                            )
            except Exception as e:  # noqa: BLE001, PERF203
                problems.add_error(e).add_error(ReindexError())
        # The next line is correct, though PyCharm gives a false-positive warning.
        # paths is a list of tuples, which the dict constructor will treat
        # as a list of key-values pairs, which is the intended behavior.
        self._pending = dict(sorted(paths, key=lambda item: item.path))  # type: ignore[attr-defined]
        if problems:
            return Err(problems)
        return Ok()

    def _today_dir(self) -> Path:
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        return (
            self._base_dir
            / datetime.datetime(
                now.year, now.month, now.day, tzinfo=now.tzinfo
            ).isoformat()
        )

    def _roll_curr_dir(self) -> None:
        today_dir = self._today_dir()
        if today_dir != self._curr_dir:
            self._curr_dir = today_dir
        if not self._curr_dir.exists():
            self._curr_dir.mkdir(parents=True, exist_ok=True)

    @classmethod
    def _make_name(cls, dt: datetime.datetime, uid: str) -> str:
        return f"{dt.isoformat()}.uid[{uid}].json"

    @classmethod
    def _persisted_item_from_file_path(cls, filepath: Path) -> Optional[_PersistedItem]:
        item = None

        try:
            match = cls.FILENAME_RGX.match(filepath.name)
            if match and cls._is_iso_parseable(match.group("dt")):
                item = _PersistedItem(match.group("uid"), filepath)
        except:  # pragma: no cover  # noqa: E722, S110
            pass
        return item

    @classmethod
    def _is_iso_parseable(cls, s: str | Path) -> bool:
        try:
            if isinstance(s, Path):
                s = s.name
            return isinstance(datetime.datetime.fromisoformat(s), datetime.datetime)
        except:  # noqa: E722
            return False



import abc
import asyncio
import functools
import logging
import os
import signal
import threading
import traceback
from dataclasses import dataclass, field
from pathlib import Path
from types import ModuleType
from typing import Any, Optional, Self, Sequence, Type

import rich
import typer
from aiohttp.typedefs import Handler as HTTPHandler
from gwproto import HardwareLayout, Message, ShNode
from gwproto.messages import EventT
from gwproto.named_types.web_server_gt import WebServerGt
from paho.mqtt.client import MQTTMessageInfo
from result import Result

from gwproactor import actors
from gwproactor.actors.actor import PrimeActor
from gwproactor.callbacks import ProactorCallbackInterface
from gwproactor.codecs import CodecFactory
from gwproactor.config import MQTTClient, Paths
from gwproactor.config.app_settings import AppSettings
from gwproactor.config.links import LinkSettings
from gwproactor.config.proactor_config import ProactorConfig, ProactorName
from gwproactor.external_watchdog import ExternalWatchdogCommandBuilder
from gwproactor.links.link_settings import LinkConfig
from gwproactor.links.mqtt import QOS
from gwproactor.logger import ProactorLogger
from gwproactor.logging_setup import setup_logging
from gwproactor.message import InternalShutdownMessage
from gwproactor.persister import PersisterInterface, StubPersister
from gwproactor.proactor_implementation import Proactor
from gwproactor.proactor_interface import (
    ActorInterface,
    AppInterface,
    CommunicatorInterface,
    IOLoopInterface,
    T,
)
from gwproactor.stats import ProactorStats


class SignalException(Exception):
    signalnum: int

    def __init__(self, signalnum: int, msg: str = "") -> None:
        super().__init__(msg)
        self.signalnum = signalnum

    def __str__(self) -> str:
        s = self.__class__.__name__
        super_str = super().__str__()
        if super_str:
            s += f" <{super_str}>"
        enum_str = (
            str(signal.Signals(self.signalnum).name)
            if self.signalnum in signal.Signals
            else "Unknown signal"
        )
        s += f"  {enum_str}  ({self.signalnum})"
        return s


@dataclass
class SubTypes:
    proactor_type: type[Proactor] = Proactor
    prime_actor_type: Optional[type[PrimeActor]] = None
    app_settings_type: type[AppSettings] = AppSettings
    actors_module: Optional[ModuleType] = None


@dataclass
class ActorConfig:
    node: ShNode
    constructor_args: dict[str, Any] = field(default_factory=dict)


class App(AppInterface):
    _settings: AppSettings
    config: ProactorConfig
    links: dict[str, LinkSettings]
    sub_types: SubTypes
    _codec_factory: CodecFactory
    _proactor: Optional[Proactor] = None
    _prime_actor: Optional[PrimeActor] = None

    def __init__(  # noqa: PLR0913
        self,
        *,
        paths_name: Optional[str] = None,
        paths: Optional[Paths] = None,
        app_settings: Optional[AppSettings] = None,
        codec_factory: Optional[CodecFactory] = None,
        sub_types: Optional[SubTypes] = None,
        env_file: Optional[str | Path] = None,
    ) -> None:
        self.sub_types = self.make_subtypes() if sub_types is None else sub_types
        self._settings = self.get_settings(
            paths_name=paths_name,
            paths=paths,
            settings=app_settings,
            env_file=env_file,
            settings_type=self.sub_types.app_settings_type,
        )
        if codec_factory is None:
            if self.sub_types.prime_actor_type is not None:
                codec_factory = self.sub_types.prime_actor_type.get_codec_factory()
            else:
                codec_factory = CodecFactory()
        self.codec_factory = CodecFactory() if codec_factory is None else codec_factory
        self.config = self._make_proactor_config()
        self.links = self._get_link_settings(
            name=self.config.name,
            layout=self.config.layout,
            brokers=self.settings.brokers(),
        )

    @property
    def settings(self) -> AppSettings:
        return self._settings

    @property
    def raw_proactor(self) -> Optional[Proactor]:
        return self._proactor

    @raw_proactor.setter
    def raw_proactor(self, proactor: Optional[Proactor]) -> None:
        self._proactor = proactor

    @property
    def proactor(self) -> Proactor:
        if self._proactor is None:
            raise ValueError("proactor accessed before it was initialized")
        return self._proactor

    @property
    def has_prime_actor(self) -> bool:
        return self._prime_actor is not None

    @property
    def prime_actor(self) -> PrimeActor:
        if self._prime_actor is None:
            raise ValueError("prime_actor accessed before it was initialized")
        return self._prime_actor

    @classmethod
    def make_subtypes(cls) -> SubTypes:
        return SubTypes(
            proactor_type=cls.proactor_type(),
            app_settings_type=cls.app_settings_type(),
            prime_actor_type=cls.prime_actor_type(),
            actors_module=cls.actors_module(),
        )

    @classmethod
    def proactor_type(cls) -> type[Proactor]:
        return Proactor

    @classmethod
    def app_settings_type(cls) -> type[AppSettings]:
        return AppSettings

    @classmethod
    def prime_actor_type(cls) -> Optional[type[PrimeActor]]:
        return None

    @classmethod
    def actors_module(cls) -> Optional[ModuleType]:
        return actors

    @classmethod
    def paths_name(cls) -> Optional[str]:
        return None

    def _make_proactor_config(
        self,
    ) -> ProactorConfig:
        layout = self._load_hardware_layout(self.settings.paths.hardware_layout)
        name = self._get_name(layout)
        return ProactorConfig(
            name=name,
            settings=self.settings,
            event_persister=self._make_persister(self.settings),
            hardware_layout=layout,
        )

    def _instantiate_proactor(self) -> Proactor:
        return self.sub_types.proactor_type(self.config)

    def instantiate(self) -> Self:
        self.raw_proactor = self._instantiate_proactor()
        self._connect_links(self.proactor)
        if self.sub_types.prime_actor_type is not None:
            self._prime_actor = self.sub_types.prime_actor_type(
                self.config.name.short_name,
                self.proactor,
            )
        self._load_actors()
        self.proactor.links.log_subscriptions("construction")
        return self

    def run_in_thread(self, *, daemon: bool = True) -> threading.Thread:
        if self.proactor is None:
            raise ValueError("ERROR. Call instantiate() before run_in_thread()")
        return self.proactor.run_in_thread(daemon=daemon)

    @abc.abstractmethod
    def _get_name(self, layout: HardwareLayout) -> ProactorName:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_link_settings(
        self,
        name: ProactorName,
        layout: HardwareLayout,
        brokers: dict[str, MQTTClient],
    ) -> dict[str, LinkSettings]:
        raise NotImplementedError

    def _connect_links(self, proactor: Proactor) -> None:
        for link_name, link_settings in self.links.items():
            if link_settings.enabled:
                proactor.links.add_mqtt_link(
                    LinkConfig(
                        client_name=link_name,
                        gnode_name=link_settings.peer_long_name,
                        spaceheat_name=link_settings.peer_short_name,
                        subscription_name=link_settings.link_subscription_short_name,
                        mqtt=self.settings.broker(link_name),
                        codec=self.codec_factory.get_codec(
                            link_name=link_name,
                            link=link_settings,
                            proactor_name=proactor.name_object,
                            layout=proactor.hardware_layout,
                        ),
                        upstream=link_settings.upstream,
                        downstream=link_settings.downstream,
                    ),
                )
        if self.settings.logging.paho_logging:
            proactor.links.enable_mqtt_loggers()

    # noinspection PyMethodMayBeStatic
    def _load_hardware_layout(self, layout_path: str | Path) -> HardwareLayout:
        return HardwareLayout.load(layout_path)

    def _make_persister(self, settings: AppSettings) -> PersisterInterface:  # noqa: ARG002, ARG003
        return StubPersister()

    def _get_actor_nodes(self) -> Sequence[ActorConfig]:
        if self._prime_actor is None or self._prime_actor.node is None:
            return []
        return [
            ActorConfig(node=node)
            for node in self.hardware_layout.nodes.values()
            if (
                node.has_actor
                and self.hardware_layout.parent_node(node) == self.prime_actor.node
            )
        ]

    def _load_actors(self) -> None:
        if self.proactor is None:
            raise ValueError("_load_actors called before proactor instantiated")
        if self.sub_types.actors_module is not None:
            for actor_config in self._get_actor_nodes():
                self.proactor.add_communicator(
                    ActorInterface.load(
                        actor_config.node.Name,
                        actor_config.node.actor_class_str,
                        self,
                        actors_module=self.sub_types.actors_module,
                        **actor_config.constructor_args,
                    )
                )

    @classmethod
    def default_env_path(cls) -> Path:
        paths_name = cls.paths_name()
        if paths_name is None:
            raise ValueError("ERROR. default_env_path() requires a non-None paths_name")
        return Paths.default_env_path(paths_name)

    @classmethod
    def get_settings(
        cls,
        paths_name: Optional[str] = None,
        paths: Optional[Paths] = None,
        settings: Optional[AppSettings] = None,
        settings_type: Optional[type[AppSettings]] = None,
        env_file: Optional[str | Path] = None,
    ) -> AppSettings:
        if settings_type is None:
            settings_type = cls.app_settings_type()
        return (
            # https://github.com/koxudaxi/pydantic-pycharm-plugin/issues/1013
            settings_type(_env_file=env_file)  # noqa
            if settings is None
            else settings.model_copy(deep=True)
        ).with_paths(name=paths_name if paths_name else cls.paths_name(), paths=paths)

    @classmethod
    def update_settings_from_command_line(  # noqa: PLR0913
        cls,
        app_settings: AppSettings,
        *,
        verbose: bool = False,
        message_summary: bool = False,
        io_loop_verbose: bool = False,
        io_loop_on_screen: bool = False,
        aiohttp_logging: bool = False,
        paho_logging: bool = False,
        **kwargs: Any,  # noqa: ARG003
    ) -> AppSettings:
        if verbose:
            app_settings.logging.base_log_level = logging.INFO
            app_settings.logging.levels.message_summary = logging.DEBUG
        elif message_summary:
            app_settings.logging.levels.message_summary = logging.INFO
        if io_loop_verbose:
            app_settings.logging.levels.io_loop = logging.DEBUG
        if io_loop_on_screen:
            app_settings.logging.io_loop.on_screen = True
        if aiohttp_logging:
            app_settings.logging.aiohttp_logging = True
        if paho_logging:
            app_settings.logging.paho_logging = True
        return app_settings

    @classmethod
    def make_app_for_cli(  # noqa: PLR0913
        cls,
        *,
        app_settings: AppSettings,
        codec_factory: Optional[CodecFactory] = None,
        sub_types: Optional[SubTypes] = None,
        env_file: Optional[str | Path] = None,
        dry_run: bool = False,
        add_screen_handler: bool = True,
    ) -> "App":
        env_file = cls.default_env_path() if not env_file else Path(env_file)
        app = cls(
            app_settings=app_settings,
            codec_factory=codec_factory,
            sub_types=sub_types,
            env_file=env_file,
        )
        env_file_debug_str = (
            f"Env file: <{env_file}>  exists: {Path(env_file).exists()}"
        )
        if dry_run:
            rich.print(env_file_debug_str)
            rich.print(app.settings)
            missing_tls_paths_ = app.settings.check_tls_paths_present(raise_error=False)
            if missing_tls_paths_:
                rich.print(missing_tls_paths_)
            rich.print("Dry run. Doing nothing.")
        else:
            app.settings.paths.mkdirs()
            setup_logging(app.settings, add_screen_handler=add_screen_handler)
            logger = logging.getLogger(
                app.settings.logging.qualified_logger_names()["lifecycle"]
            )
            logger.info("")
            logger.info(env_file_debug_str)
            logger.info("Settings:")
            logger.info(app.settings.model_dump_json(indent=2))
            rich.print(app.settings)
            app.settings.check_tls_paths_present()
            app.instantiate()
        return app

    def signal_handler(self, signum: int, _frame: Any, *, tag: str) -> None:
        try:
            try:
                enum_str = signal.Signals(signum).name
            except:  # noqa: E722
                enum_str = "Unknown Signal"
            tag_str = f"  {tag}" if tag else ""
            self.send_threadsafe(
                InternalShutdownMessage(
                    Src="SignalHandler",
                    Reason=f"Signal received: {enum_str} ({signum}){tag_str}",
                )
            )
        except BaseException as e:
            raise SignalException(
                signum, msg=f"Signal received. Exception while sending to Proactor: {e}"
            ) from e

    def install_signal_handlers(self) -> None:
        signal.signal(
            signal.SIGTERM,
            functools.partial(
                self.signal_handler,
                tag="Service stopped" if self.running_as_service() else "",
            ),
        )
        signal.signal(
            signal.SIGABRT,
            functools.partial(
                self.signal_handler,
                tag="Service failed to path SystemD watchdog"
                if self.running_as_service()
                else "",
            ),
        )
        signal.signal(
            signal.SIGINT,
            functools.partial(
                self.signal_handler,
                tag="KeyboardInterrupt",
            ),
        )

    @classmethod
    def main(  # noqa: PLR0913
        cls,
        *,
        paths_name: Optional[str] = None,
        paths: Optional[Paths] = None,
        app_settings: Optional[AppSettings] = None,
        codec_factory: Optional[CodecFactory] = None,
        sub_types: Optional[SubTypes] = None,
        env_file: Optional[str | Path] = "",
        dry_run: bool = False,
        verbose: bool = False,
        message_summary: bool = False,
        io_loop_verbose: bool = False,
        io_loop_on_screen: bool = False,
        aiohttp_logging: bool = False,
        paho_logging: bool = False,
        add_screen_handler: bool = True,
        run_in_thread: bool = False,
        return_int: bool = False,
    ) -> int:
        env_file = cls.default_env_path() if not env_file else Path(env_file)
        if app_settings is None:
            app_settings = cls.get_settings(
                paths_name=paths_name,
                paths=paths,
                env_file=env_file,
            )
        app_settings = cls.update_settings_from_command_line(
            app_settings=app_settings,
            verbose=verbose,
            message_summary=message_summary,
            io_loop_verbose=io_loop_verbose,
            io_loop_on_screen=io_loop_on_screen,
            aiohttp_logging=aiohttp_logging,
            paho_logging=paho_logging,
        )
        exception_logger: logging.Logger | logging.LoggerAdapter[logging.Logger] = (
            logging.getLogger(app_settings.logging.base_log_name)
        )
        ret = 0
        try:
            app = cls.make_app_for_cli(
                app_settings=app_settings,
                codec_factory=codec_factory,
                sub_types=sub_types,
                env_file=env_file,
                dry_run=dry_run,
                add_screen_handler=add_screen_handler,
            )
            if not dry_run:
                if run_in_thread:
                    app.run_in_thread()
                else:
                    app.install_signal_handlers()
                    try:
                        asyncio.run(app.proactor.run_forever())
                    finally:
                        app.proactor.stop()
        except SystemExit:
            ret = 1
        except KeyboardInterrupt:
            ret = 2
        except BaseException as e:  # noqa: BLE001
            ret = 3
            try:
                exception_logger.exception(
                    "ERROR in app_main. Shutting down: [%s] / [%s]",
                    e,  # noqa: TRY401
                    type(e),  # noqa: TRY401
                )
            except:  # noqa: E722
                traceback.print_exception(e)
        if not return_int:
            raise typer.Exit(code=ret)
        return ret

    @classmethod
    def print_settings(
        cls,
        *,
        env_file: Optional[str | Path] = None,
    ) -> None:
        env_file = cls.default_env_path() if not env_file else Path(env_file)
        rich.print(
            f"Env file: <{env_file}>  exists: {bool(env_file and Path(env_file).exists())}"
        )
        app = cls(env_file=env_file)
        rich.print(app.settings)
        missing_tls_paths_ = app.settings.check_tls_paths_present(raise_error=False)
        if missing_tls_paths_:
            rich.print(missing_tls_paths_)

    @classmethod
    def service_variable_name(cls) -> str:
        return f"{(cls.paths_name() or 'GWPROACTOR').upper()}_RUNNING_AS_SERVICE"

    @classmethod
    def running_as_service(cls) -> bool:
        return os.getenv(cls.service_variable_name(), "").lower() in [
            "1",
            "true",
        ]

    @property
    def name(self) -> str:
        return self.config.name.name

    def add_communicator(self, communicator: CommunicatorInterface) -> None:
        self.proactor.add_communicator(communicator)

    def get_communicator(self, name: str) -> Optional[CommunicatorInterface]:
        return self.proactor.get_communicator(name)

    def get_communicator_as_type(self, name: str, type_: Type[T]) -> Optional[T]:
        return self.proactor.get_communicator_as_type(name, type_)

    def get_communicator_names(self) -> set[str]:
        return self.proactor.get_communicator_names()

    def send(self, message: Message[Any]) -> None:
        self.proactor.send(message)

    async def await_processing(
        self, message: Message[Any]
    ) -> Result[Any, BaseException]:
        return await self.proactor.await_processing(message)

    def send_threadsafe(self, message: Message[Any]) -> None:
        self.proactor.send_threadsafe(message)

    def wait_for_processing_threadsafe(
        self, message: Message[Any]
    ) -> Result[Any, BaseException]:
        return self.proactor.wait_for_processing_threadsafe(message)

    def add_task(self, task: asyncio.Task[Any]) -> None:
        self.proactor.add_task(task)

    @property
    def async_receive_queue(self) -> Optional[asyncio.Queue[Any]]:
        return self.proactor.async_receive_queue

    @property
    def event_loop(self) -> Optional[asyncio.AbstractEventLoop]:
        return self.proactor.event_loop

    @property
    def io_loop_manager(self) -> IOLoopInterface:
        return self.proactor.io_loop_manager

    def add_web_server_config(
        self, name: str, host: str, port: int, **kwargs: Any
    ) -> None:
        self.proactor.add_web_server_config(name, host, port, **kwargs)

    def add_web_route(
        self,
        server_name: str,
        method: str,
        path: str,
        handler: HTTPHandler,
        **kwargs: Any,
    ) -> None:
        self.proactor.add_web_route(server_name, method, path, handler, **kwargs)

    def get_web_server_route_strings(self) -> dict[str, list[str]]:
        return self.proactor.get_web_server_route_strings()

    def get_web_server_configs(self) -> dict[str, WebServerGt]:
        return self.proactor.get_web_server_configs()

    def generate_event(self, event: EventT) -> Result[bool, Exception]:
        return self.proactor.generate_event(event)

    @property
    def publication_name(self) -> str:
        return self.config.name.publication_name

    @property
    def subscription_name(self) -> str:
        return self.config.name.subscription_name

    def publish_message(  # noqa: PLR0913
        self,
        link_name: str,
        message: Message[Any],
        qos: int = 0,
        context: Any = None,
        *,
        topic: str = "",
        use_link_topic: bool = False,
    ) -> MQTTMessageInfo:
        if self.proactor is None:
            raise ValueError("publish_message called before proactor instantiated")
        return self.proactor.publish_message(
            link_name=link_name,
            message=message,
            qos=qos,
            context=context,
            topic=topic,
            use_link_topic=use_link_topic,
        )

    def publish_upstream(
        self, payload: Any, qos: QOS = QOS.AtMostOnce, **message_args: Any
    ) -> MQTTMessageInfo:
        return self.proactor.publish_upstream(payload=payload, qos=qos, **message_args)

    @property
    def upstream_client(self) -> str:
        return self.proactor.upstream_client

    @property
    def downstream_client(self) -> str:
        return self.proactor.downstream_client

    @property
    def logger(self) -> ProactorLogger:
        return self.config.logger

    @property
    def stats(self) -> ProactorStats:
        return self.proactor.stats

    @property
    def hardware_layout(self) -> HardwareLayout:
        return self.config.layout

    def get_external_watchdog_builder_class(
        self,
    ) -> type[ExternalWatchdogCommandBuilder]:
        return self.proactor.get_external_watchdog_builder_class()

    def add_callbacks(self, callbacks: ProactorCallbackInterface) -> int:
        return self.proactor.add_callbacks(callbacks)

    def remove_callbacks(self, callbacks_id: int) -> None:
        self.proactor.remove_callbacks(callbacks_id)



import abc
from abc import abstractmethod
from typing import Optional

from result import Result

from gwproactor.problems import Problems

ENCODING: str = "utf-8"


class PersisterInterface(abc.ABC):
    @abstractmethod
    def persist(self, uid: str, content: bytes) -> Result[bool, Problems]:
        """Persist content, indexed by uid"""

    @abstractmethod
    def clear(self, uid: str) -> Result[bool, Problems]:
        """Delete content persisted for uid. It is error to clear a uid which is not currently persisted."""

    @abstractmethod
    def pending_ids(self) -> list[str]:
        """Get list of pending (persisted and not cleared) uids"""

    @property
    @abstractmethod
    def num_pending(self) -> int:
        """Get number of pending uids"""

    @property
    @abstractmethod
    def curr_bytes(self) -> int:
        """Return number of bytes used to store events, if known."""

    @abstractmethod
    def __contains__(self, uid: str) -> bool:
        """Check whether a uid is pending"""

    @abstractmethod
    def retrieve(self, uid: str) -> Result[Optional[bytes], Problems]:
        """Load and return persisted content for uid"""

    @abstractmethod
    def reindex(self) -> Result[Optional[bool], Problems]:
        """Re-created pending index from persisted storage"""

    @property
    @abstractmethod
    def num_persists(self) -> int:
        """Total number of calls to persist() since construction."""

    @property
    @abstractmethod
    def num_retrieves(self) -> int:
        """Total number of calls to retrieve() since construction."""

    @property
    @abstractmethod
    def num_clears(self) -> int:
        """Total number of calls to clear() since construction."""



import typer

from gwproactor_test.dummies.tree.scada1 import DummyScada1App

app = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    rich_markup_mode="rich",
    help="GridWorks Dummy Scada1",
)


@app.command()
def run(
    env_file: str = "",
    dry_run: bool = False,
    verbose: bool = False,
    message_summary: bool = False,
    aiohttp_verbose: bool = False,
    io_loop_verbose: bool = False,
    io_loop_on_screen: bool = False,
) -> None:
    DummyScada1App.main(
        env_file=env_file,
        dry_run=dry_run,
        verbose=verbose,
        message_summary=message_summary,
        io_loop_verbose=io_loop_verbose,
        io_loop_on_screen=io_loop_on_screen,
        aiohttp_logging=aiohttp_verbose,
    )


@app.command()
def config(
    env_file: str = "",
) -> None:
    DummyScada1App.print_settings(env_file=env_file)


@app.callback()
def _main() -> None: ...


if __name__ == "__main__":
    app()



from pathlib import Path
from typing import Any, Optional, Self, TypeVar

from pydantic import Field, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from gwproactor.config.logging import LoggingSettings
from gwproactor.config.mqtt import MQTTClient
from gwproactor.config.paths import Paths
from gwproactor.config.proactor_settings import ProactorSettings

FieldT = TypeVar("FieldT")


class AppSettings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="PROACTOR_APP_",
        env_nested_delimiter="__",
        nested_model_default_partial_update=True,
        extra="ignore",
    )
    paths: Paths = Field(default_factory=Paths, validate_default=True)
    logging: LoggingSettings = LoggingSettings()
    proactor: ProactorSettings = ProactorSettings()

    def brokers(self) -> dict[str, MQTTClient]:
        return self._fields_of_type(MQTTClient)

    def broker(self, name: str) -> MQTTClient:
        return self._field_as_type(name, MQTTClient)

    def update_paths_name(self, name: str | Path) -> Self:
        self.paths = Paths(
            name=name, **self.paths.model_dump(exclude={"name"}, exclude_unset=True)
        )
        self.update_tls_paths()
        return self

    def with_paths(
        self,
        *,
        paths: Optional[Paths] = None,
        exclude_unset: bool = True,
        exclude_defaults: bool = True,
        **kwargs: Any,
    ) -> Self:
        if paths is None:
            paths = self.paths
        self.paths = paths.duplicate(
            exclude=set(kwargs.keys()),
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            **kwargs,
        )
        self.update_tls_paths()
        return self

    def with_paths_name(self, name: str) -> Self:
        return self.with_paths(name=name)

    @classmethod
    def update_paths_name_before_validator(
        cls, values: dict[str, Any], name: str
    ) -> dict[str, Any]:
        """Update paths member with a new 'name' attribute, e.g., a name known by a derived class.

        This may be called in a mode="before" root validator of a derived class.
        """
        if "paths" not in values:
            values["paths"] = Paths(name=name)
        elif isinstance(values["paths"], Paths):
            if "name" not in values["paths"].model_fields_set:
                values["paths"] = values["paths"].duplicate(name=name)
        elif "name" not in values["paths"]:
            values["paths"]["name"] = name
        return values

    @model_validator(mode="after")
    def post_root_validator(self) -> Self:
        """Update unset paths of any member MQTTClient's TLS paths based on ProactorSettings 'paths' member."""
        if not isinstance(self.paths, Paths):
            raise ValueError(  # noqa: TRY004
                f"ERROR. 'paths' member must be instance of Paths. Got: {type(self.paths)}"
            )
        self.update_tls_paths()
        return self

    def update_tls_paths(self) -> Self:
        for client_name, client in self.brokers().items():
            client.update_tls_paths(Path(self.paths.certs_dir), client_name)
        return self

    def uses_tls(self) -> bool:
        return any(client.tls.use_tls for client in self.brokers().values())

    def check_tls_paths_present(self, *, raise_error: bool = True) -> str:
        missing_str = ""
        for broker_name, broker in self.brokers().items():
            if broker.tls.use_tls:
                missing_paths = broker.tls.paths.missing_paths()
                if missing_paths:
                    missing_str += f"broker {broker_name}\n"
                    for path_name, path in missing_paths:
                        missing_str += f"  {path_name:20s}  {path}\n"
        if missing_str:
            error_str = f"ERROR. TLS usage requested but the following files are missing:\n{missing_str}"
            if raise_error:
                raise ValueError(error_str)
        else:
            error_str = ""
        return error_str

    def _fields_of_type(self, field_type: type[FieldT]) -> dict[str, FieldT]:
        return {
            field_name: getattr(self, field_name)
            for field_name in self.__pydantic_fields__
            if isinstance(getattr(self, field_name, None), field_type)
        }

    def _field_as_type(self, field_name: str, field_type: type[FieldT]) -> FieldT:
        if field_name not in self.__pydantic_fields__:
            raise ValueError(
                f"ERROR. AppSettings field <{field_name}> not present. Was expected as type {field_type}"
            )
        field_value = getattr(self, field_name, None)
        if not isinstance(field_value, field_type):
            raise ValueError(  # noqa: TRY004
                f"ERROR. Field name <{field_type}> is present (type: {type(field_value)} "
                f"but is not an instance of {field_type}."
            )
        return field_value



"""Scada implementation"""

import typing
from typing import Any, Optional

from gwproto import HardwareLayout, Message
from gwproto.messages import EventBase

from gwproactor import App, AppSettings
from gwproactor.actors.actor import PrimeActor
from gwproactor.config import MQTTClient
from gwproactor.config.links import LinkSettings
from gwproactor.config.proactor_config import ProactorName
from gwproactor.message import MQTTReceiptPayload
from gwproactor.persister import (
    PersisterInterface,
    TimedRollingFilePersister,
)
from gwproactor_test.dummies import DUMMY_SCADA1_NAME
from gwproactor_test.dummies.names import DUMMY_ATN_NAME


class DummyAtn(PrimeActor):
    def process_mqtt_message(
        self, mqtt_client_message: Message[MQTTReceiptPayload], decoded: Message[Any]
    ) -> None:
        self.services.logger.path(
            f"++{self.name}.process_mqtt_message %s",
            mqtt_client_message.Payload.message.topic,
        )
        path_dbg = 0
        self.services.stats.add_message(decoded)
        match decoded.Payload:
            case EventBase():
                path_dbg |= 0x00000001
                self.services.generate_event(decoded.Payload)
            case _:
                path_dbg |= 0x00000002
        self.services.logger.path(
            f"--{self.name}.process_mqtt_message  path:0x%08X", path_dbg
        )


class DummyAtnSettings(AppSettings):
    dummy_scada1: MQTTClient = MQTTClient()


class DummyAtnApp(App):
    SCADA1_LINK: str = DUMMY_SCADA1_NAME

    @classmethod
    def app_settings_type(cls) -> type[AppSettings]:
        return DummyAtnSettings

    @classmethod
    def prime_actor_type(cls) -> type[DummyAtn]:
        return DummyAtn

    @property
    def prime_actor(self) -> DummyAtn:
        return typing.cast(DummyAtn, self._prime_actor)

    @classmethod
    def paths_name(cls) -> Optional[str]:
        return DUMMY_ATN_NAME

    def _get_name(self, layout: HardwareLayout) -> ProactorName:
        return ProactorName(
            long_name=layout.atn_g_node_alias,
            short_name="a",
        )

    def _get_link_settings(
        self,
        name: ProactorName,  # noqa: ARG002
        layout: HardwareLayout,
        brokers: dict[str, MQTTClient],  # noqa: ARG002
    ) -> dict[str, LinkSettings]:
        return {
            self.SCADA1_LINK: LinkSettings(
                broker_name=self.SCADA1_LINK,
                peer_long_name=layout.scada_g_node_alias,
                peer_short_name="s",
                downstream=True,
            )
        }

    @classmethod
    def _make_persister(cls, settings: AppSettings) -> PersisterInterface:
        return TimedRollingFilePersister(settings.paths.event_dir)



import textwrap
import typing
from typing import Any, Optional, Self

from gwproactor import AppSettings, setup_logging
from gwproactor.app import App
from gwproactor.config import MQTTClient
from gwproactor_test.dummies.tree.atn import DummyAtnApp
from gwproactor_test.dummies.tree.scada1 import DummyScada1App
from gwproactor_test.dummies.tree.scada2 import DummyScada2App
from gwproactor_test.instrumented_proactor import InstrumentedProactor
from gwproactor_test.live_test_helper import (
    LiveTest,
    get_option_value,
)
from gwproactor_test.logger_guard import LoggerGuards


class TreeLiveTest(LiveTest):
    _child2_app: App
    child2_verbose: bool = False
    child2_on_screen: bool = False
    child2_logger_guards: LoggerGuards

    def __init__(
        self,
        *,
        add_child1: bool = False,
        start_child1: bool = False,
        child1_verbose: Optional[bool] = None,
        child2_app_settings: Optional[AppSettings] = None,
        child2_verbose: Optional[bool] = None,
        add_child2: bool = False,
        start_child2: bool = False,
        child2_on_screen: Optional[bool] = None,
        **kwargs: Any,
    ) -> None:
        kwargs["add_child"] = add_child1 or kwargs.get("add_child", False)
        kwargs["start_child"] = start_child1 or kwargs.get("start_child", False)
        if child1_verbose is None:
            kwargs["child_verbose"] = get_option_value(
                parameter_value=child1_verbose,
                option_name="--child1-verbose",
                request=kwargs.get("request"),
            )
        kwargs["request"] = kwargs.get("request")
        super().__init__(**kwargs)
        self.child2_verbose = get_option_value(
            parameter_value=child2_verbose,
            option_name="--child2-verbose",
            request=kwargs.get("request"),
        )
        self.child2_on_screen = get_option_value(
            parameter_value=child2_on_screen,
            option_name="--child2-on-screen",
            request=kwargs.get("request"),
        )
        self._child2_app = self._make_app(
            self.child2_app_type(), child2_app_settings, app_verbose=self.child2_verbose
        )
        self.setup_child2_logging()
        if add_child2 or start_child2:
            self.add_child2()
            if start_child2:
                self.start_child2()

    @classmethod
    def child_app_type(cls) -> type[App]:
        return DummyScada1App

    @property
    def child_app(self) -> DummyScada1App:
        return typing.cast(DummyScada1App, self._child_app)

    @property
    def child1_app(self) -> DummyScada1App:
        return self.child_app

    @classmethod
    def child2_app_type(cls) -> type[App]:
        return DummyScada2App

    @property
    def child2_app(self) -> DummyScada2App:
        return typing.cast(DummyScada2App, self._child2_app)

    @classmethod
    def parent_app_type(cls) -> type[App]:
        return DummyAtnApp

    @property
    def parent_app(self) -> DummyAtnApp:
        return typing.cast(DummyAtnApp, self._parent_app)

    @property
    def child1(self) -> InstrumentedProactor:
        return self.child

    def add_child1(self) -> Self:
        return self.add_child()

    def start_child1(
        self,
    ) -> Self:
        return self.start_child()

    def remove_child1(
        self,
    ) -> Self:
        return self.remove_child()

    @property
    def child2(self) -> InstrumentedProactor:
        if self.child2_app.proactor is None:
            raise RuntimeError(
                "ERROR. CommTestHelper.child accessed before creating child."
                "pass add_child=True to CommTestHelper constructor or call "
                "CommTestHelper.add_child()"
            )
        return typing.cast(InstrumentedProactor, self.child2_app.proactor)

    def add_child2(
        self,
    ) -> Self:
        self.child2_app.instantiate()
        return self

    def start_child2(
        self,
    ) -> Self:
        if self.child2_app.raw_proactor is None:
            self.add_child2()
        return self.start_proactor(self.child2)

    def remove_child2(
        self,
    ) -> Self:
        self.child2_app.raw_proactor = None
        return self

    def _get_child2_clients_supporting_tls(self) -> list[MQTTClient]:
        return self._get_clients_supporting_tls(self.child2_app.config.settings)

    def set_use_tls(self, use_tls: bool) -> None:
        super().set_use_tls(use_tls)
        self._set_settings_use_tls(use_tls, self._get_child2_clients_supporting_tls())

    def setup_child2_logging(self) -> None:
        self.child2_app.config.settings.paths.mkdirs(parents=True)
        errors: list[Exception] = []
        self.logger_guards.add_loggers(
            list(
                self.child2_app.config.settings.logging.qualified_logger_names().values()
            )
        )
        setup_logging(
            self.child2_app.config.settings,
            errors=errors,
            add_screen_handler=self.child2_on_screen,
            root_gets_handlers=False,
        )
        assert not errors

    def get_proactors(self) -> list[InstrumentedProactor]:
        proactors = super().get_proactors()
        if self.child2_app.raw_proactor is not None:
            proactors.append(self.child2)
        return proactors

    def get_log_path_str(self, exc: BaseException) -> str:
        return (
            f"CommTestHelper caught error {exc}.\n"
            "Working log dirs:"
            f"\n\t[{self.child_app.config.settings.paths.log_dir}]"
            f"\n\t[{self.parent_app.config.settings.paths.log_dir}]"
        )

    def summary_str(self) -> str:
        s = ""
        if self.child_app.raw_proactor is None:
            s += "SCADA1: None\n"
        else:
            s += "SCADA1:\n"
            s += textwrap.indent(self.child1.summary_str(), "    ") + "\n"
        if self.child2_app.raw_proactor is None:
            s += "SCADA2: None\n"
        else:
            s += "SCADA2:\n"
            s += textwrap.indent(self.child2.summary_str(), "    ") + "\n"
        if self.parent_app.raw_proactor is None:
            s += "ATN: None\n"
        else:
            s += "ATN:\n"
            s += textwrap.indent(self.parent.summary_str(), "    ") + "\n"
        return s



import secrets
from typing import Any, ClassVar, Optional, Sequence

from gwproto import HardwareLayout, Message, MQTTCodec, create_message_model
from gwproto.messages import AnyEvent
from pydantic import ValidationError
from pydantic_core import ErrorDetails

from gwproactor.config.links import LinkSettings
from gwproactor.config.proactor_config import ProactorName


class ProactorCodec(MQTTCodec):
    DEFAULT_MESSAGE_MODULES: ClassVar[list[str]] = [
        "gwproto.messages",
        "gwproactor.message",
    ]
    src_name: str
    dst_name: str

    def __init__(
        self,
        *,
        src_name: str = "",
        dst_name: str = "",
        model_name: str = "",
        module_names: Optional[Sequence[str]] = None,
        use_default_modules: bool = True,
    ) -> None:
        self.src_name = src_name
        self.dst_name = dst_name
        super().__init__(
            self.create_message_model(
                model_name=model_name,
                module_names=module_names,
                use_default_modules=use_default_modules,
            )
        )

    @classmethod
    def create_message_model(
        cls,
        *,
        model_name: str = "",
        module_names: Optional[Sequence[str]] = None,
        use_default_modules: bool = True,
    ) -> type[Message[Any]]:
        module_names_used = []
        if use_default_modules:
            module_names_used.extend(cls.DEFAULT_MESSAGE_MODULES)
        if module_names is not None:
            module_names_used.extend(module_names)
        model_name = (
            model_name if model_name else "ProactorCodec-" + secrets.token_hex(4)
        )
        return create_message_model(
            model_name=model_name, module_names=module_names_used
        )

    def validate_source_and_destination(self, src: str, dst: str) -> None:
        if (self.src_name and src != self.src_name) or (
            self.dst_name and dst != self.dst_name
        ):
            raise ValueError(
                "ERROR validating src and/or dst\n"
                f"  exp: {self.src_name} -> {self.dst_name}\n"
                f"  got: {src} -> {dst}"
            )

    @classmethod
    def _may_be_event(cls, details: ErrorDetails) -> bool:
        loc: Sequence[str | int] = details.get("loc", [])
        return (
            len(loc) >= 2  # noqa: PLR2004
            and loc[0] == "Payload"
            and isinstance(loc[1], str)
            and loc[1].startswith("gridworks.event")
        )

    @classmethod
    def get_unrecognized_payload_error(
        cls, e: ValidationError
    ) -> Optional[ErrorDetails]:
        super_error = super().get_unrecognized_payload_error(e)
        if super_error is None:
            for error in e.errors():
                if cls._may_be_event(error):
                    return error
        return super_error

    def handle_unrecognized_payload(  # noqa
        self, payload: bytes, e: ValidationError, details: ErrorDetails
    ) -> Message[Any]:
        if self._may_be_event(details):
            try:
                return Message[AnyEvent].model_validate_json(payload)
            except ValidationError as e2:
                raise e2 from e
        return super().handle_unrecognized_payload(
            payload=payload, e=e, details=details
        )


class CodecFactory:
    # noinspection PyMethodMayBeStatic
    def get_codec(
        self,
        link_name: str,  # noqa: ARG002
        link: LinkSettings,
        proactor_name: ProactorName,
        layout: HardwareLayout,  # noqa: ARG002
    ) -> MQTTCodec:
        return ProactorCodec(
            src_name=link.peer_long_name,
            dst_name=proactor_name.short_name,
            model_name=link.codec.message_model_name,
            module_names=link.codec.message_modules,
            use_default_modules=link.codec.use_default_message_modules,
        )



# ruff: noqa: ERA001

import datetime
from pathlib import Path

from result import Err, Ok, Result

from gwproactor.persister.exceptions import PersisterError, WriteFailed
from gwproactor.persister.stub import StubPersister
from gwproactor.problems import Problems


class SimpleDirectoryWriter(StubPersister):
    _base_dir: Path

    def __init__(
        self,
        base_dir: Path | str,
    ) -> None:
        self._base_dir = Path(base_dir).resolve()

    @classmethod
    def _make_name(cls, dt: datetime.datetime, uid: str) -> str:
        return f"{dt.isoformat()}.uid[{uid}].json"

    def persist(self, uid: str, content: bytes) -> Result[bool, Problems]:
        self._num_persists += 1
        # from gwproto.messages import AnyEvent
        # event = AnyEvent.model_validate_json(content.decode())
        # print(
        #     f"{self._num_persists:3d}  "
        #     f"{event.TypeName:45s}  "
        #     f"{event.Src:35s}  "
        #     f"{event.MessageId[:8]}"
        # )
        problems = Problems()
        try:
            if not self._base_dir.exists():
                self._base_dir.mkdir(parents=True, exist_ok=True)
            path = self._base_dir / self._make_name(
                datetime.datetime.now(tz=datetime.timezone.utc), uid
            )
            try:
                with path.open("wb") as f:
                    f.write(content)
            except Exception as e:  # pragma: no cover  # noqa: BLE001
                return Err(
                    problems.add_error(e).add_error(
                        WriteFailed("Open or write failed", uid=uid, path=path)
                    )
                )
        except Exception as e:  # noqa: BLE001
            return Err(
                problems.add_error(e).add_error(
                    PersisterError("Unexpected error", uid=uid)
                )
            )
        if problems:
            return Err(problems)
        return Ok()



import typing
from typing import Any, Self

from pydantic import model_validator
from pydantic_settings import SettingsConfigDict

from gwproactor import AppSettings
from gwproactor_test.dummies.names import DUMMY_ADMIN_NAME, DUMMY_ADMIN_SHORT_NAME
from gwproactor_test.dummies.tree.link_settings import TreeLinkSettings


class AdminLinkSettings(TreeLinkSettings):
    DUMMY_ADMIN_CLIENT_NAME: typing.ClassVar[str] = "dummy_admin"
    DUMMY_ADMIN_PATHS_NAME: typing.ClassVar[str] = "dummy_admin"

    def __init__(self, **kwargs: Any) -> None:
        kwargs["client_name"] = kwargs.get("client_name", self.DUMMY_ADMIN_CLIENT_NAME)
        kwargs["long_name"] = kwargs.get("long_name", DUMMY_ADMIN_NAME)
        kwargs["short_name"] = kwargs.get("short_name", DUMMY_ADMIN_SHORT_NAME)
        super().__init__(**kwargs)


class DummyAdminSettings(AppSettings):
    DEFAULT_TARGET: typing.ClassVar[str] = "d1.isone.ct.newhaven.orange1.scada"
    target_gnode: str = DEFAULT_TARGET
    link: AdminLinkSettings = AdminLinkSettings()
    model_config = SettingsConfigDict(env_prefix="GWADMIN_", env_nested_delimiter="__")

    @model_validator(mode="before")
    @classmethod
    def pre_root_validator(cls, values: dict[str, Any]) -> dict[str, Any]:
        return cls.update_paths_name_before_validator(
            values, AdminLinkSettings.DUMMY_ADMIN_PATHS_NAME
        )

    @model_validator(mode="after")
    def validate_(self) -> Self:
        self.link.update_tls_paths(self.paths.certs_dir, self.link.client_name)
        return self



import asyncio
import json
from dataclasses import asdict, dataclass, field
from typing import Any, Optional, Tuple

from gwproto import Message, MQTTCodec, MQTTTopic
from gwproto.messages import (
    Ack,
    CommEvent,
    EventBase,
    EventT,
    MQTTConnectEvent,
    MQTTDisconnectEvent,
    MQTTFullySubscribedEvent,
    PeerActiveEvent,
    PingMessage,
    ProblemEvent,
    ResponseTimeoutEvent,
    StartupEvent,
)
from paho.mqtt.client import MQTTMessageInfo
from result import Err, Ok, Result

from gwproactor.config import ProactorSettings
from gwproactor.links import AckWaitInfo
from gwproactor.links.acks import AckManager, AckTimerCallback
from gwproactor.links.link_settings import LinkConfig
from gwproactor.links.link_state import (
    InvalidCommStateInput,
    LinkState,
    LinkStates,
    StateName,
    Transition,
)
from gwproactor.links.message_times import LinkMessageTimes, MessageTimes
from gwproactor.links.mqtt import QOS, MQTTClients, MQTTClientWrapper
from gwproactor.links.reuploads import Reuploads
from gwproactor.links.timer_interface import TimerManagerInterface
from gwproactor.logger import LoggerOrAdapter, ProactorLogger
from gwproactor.message import (
    MQTTConnectFailPayload,
    MQTTConnectPayload,
    MQTTDisconnectPayload,
    MQTTReceiptPayload,
    MQTTSubackPayload,
)
from gwproactor.persister import (
    ByteDecodingError,
    DecodingError,
    FileEmptyWarning,
    JSONDecodingError,
    PersisterInterface,
    UIDMissingWarning,
)
from gwproactor.persister.interface import ENCODING as PERSISTER_ENCODING
from gwproactor.problems import Problems
from gwproactor.stats import ProactorStats


@dataclass
class LinkManagerTransition(Transition):
    canceled_acks: list[AckWaitInfo] = field(default_factory=list)


class LinkManager:
    PERSISTER_ENCODING = PERSISTER_ENCODING
    publication_name: str
    subscription_name: str
    _settings: ProactorSettings
    _logger: ProactorLogger
    _stats: ProactorStats
    _event_persister: PersisterInterface
    _reuploads: Reuploads
    _mqtt_clients: MQTTClients
    _mqtt_codecs: dict[str, MQTTCodec]
    _states: LinkStates
    _message_times: MessageTimes
    _acks: AckManager
    _in_flight_events: dict[str, EventBase]

    def __init__(  # noqa: PLR0913, PLR0917
        self,
        publication_name: str,
        subscription_name: str,
        settings: ProactorSettings,
        logger: ProactorLogger,
        stats: ProactorStats,
        event_persister: PersisterInterface,
        timer_manager: TimerManagerInterface,
        ack_timeout_callback: AckTimerCallback,
    ) -> None:
        self.publication_name = publication_name
        self.subscription_name = subscription_name
        self._settings = settings
        self._logger = logger
        self._stats = stats
        self._in_flight_events = {}
        self._event_persister = event_persister
        self._reuploads = Reuploads(
            self._logger,
            self._settings.num_initial_event_reuploads,
        )
        self._mqtt_clients = MQTTClients()
        self._mqtt_codecs = {}
        self._states = LinkStates()
        self._message_times = MessageTimes()
        self._acks = AckManager(
            timer_manager, ack_timeout_callback, delay=settings.ack_timeout_seconds
        )

    @property
    def upstream_client(self) -> str:
        return self._mqtt_clients.upstream_client

    @property
    def downstream_client(self) -> str:
        return self._mqtt_clients.downstream_client

    @property
    def num_pending(self) -> int:
        return self._event_persister.num_pending

    @property
    def num_reupload_pending(self) -> int:
        return self._reuploads.num_reupload_pending

    @property
    def num_reuploaded_unacked(self) -> int:
        return self._reuploads.num_reuploaded_unacked

    @property
    def ack_manager(self) -> AckManager:
        return self._acks

    def topic_dst(self, client_name: str) -> str:
        return self._mqtt_clients.topic_dst(client_name)

    def reuploading(self) -> bool:
        return self._reuploads.reuploading()

    def num_acks(self, link_name: str) -> int:
        return self._acks.num_acks(link_name)

    def subscribed(self, link_name: str) -> bool:
        return self._mqtt_clients.subscribed(link_name)

    def mqtt_clients(self) -> MQTTClients:
        return self._mqtt_clients

    def mqtt_client_wrapper(self, client_name: str) -> MQTTClientWrapper:
        return self._mqtt_clients.client_wrapper(client_name)

    def enable_mqtt_loggers(self, logger: Optional[LoggerOrAdapter] = None) -> None:
        self._mqtt_clients.enable_loggers(logger)

    def disable_mqtt_loggers(self) -> None:
        self._mqtt_clients.disable_loggers()

    def decoder(self, link_name: str) -> Optional[MQTTCodec]:
        return self._mqtt_codecs.get(link_name, None)

    def decode(self, link_name: str, topic: str, payload: bytes) -> Message[Any]:
        return self._mqtt_codecs[link_name].decode(topic, payload)

    def link(self, name: str) -> Optional[LinkState]:
        return self._states.link(name)

    def link_state(self, name: str) -> Optional[StateName]:
        return self._states.link_state(name)

    def link_names(self) -> list[str]:
        return self._states.link_names()

    def __contains__(self, name: str) -> bool:
        return name in self._states

    def __getitem__(self, name: str) -> LinkState:
        return self._states[name]

    def get_message_times(self, link_name: str) -> LinkMessageTimes:
        return self._message_times.get_copy(link_name)

    def stopped(self, name: str) -> bool:
        return self._states.stopped(name)

    def add_mqtt_link(self, settings: LinkConfig) -> None:
        self._mqtt_clients.add_client(settings)
        self._mqtt_codecs[settings.client_name] = settings.codec
        self._states.add(settings.client_name)
        self._message_times.add_link(settings.client_name)
        self._stats.add_link(settings.client_name)
        self.subscribe(
            client=settings.client_name,
            topic=settings.subscription_topic(
                settings.subscription_name
                if settings.subscription_name
                else self.subscription_name
            ),
            qos=QOS.AtMostOnce,
        )

    def subscribe(self, client: str, topic: str, qos: int) -> Tuple[int, Optional[int]]:
        return self._mqtt_clients.subscribe(client, topic, qos)

    def subscription_str(self, tag: str = "") -> str:
        tag_str = f"[{tag}]" if tag else ""
        s = f"\nSubscription info for <{self.publication_name}> {tag_str}\n"
        for client_name in self._mqtt_clients.clients:
            client = self._mqtt_clients.client_wrapper(client_name)
            s += f"  Client name: <{client_name}>  topic_dst: <{client.topic_dst}>\n"
            publish_topic = MQTTTopic.encode(
                envelope_type=Message.type_name(),
                src=self.publication_name,
                dst=client.topic_dst,
                message_type="SOME_MESSAGE_TYPE",
            )
            s += f"    Publish to: {publish_topic}\n"
            s += "    Subscriptions:\n"
            for subscription in client.subscription_items():
                s += f"      [{subscription}]\n"
        return s

    def log_subscriptions(self, tag: str = "") -> None:
        if self._logger.lifecycle_enabled:
            self._logger.lifecycle(self.subscription_str(tag=tag))

    def get_reuploads_str(self, verbose: bool = True, num_events: int = 5) -> str:  # noqa: FBT001, FBT002
        return self._reuploads.get_str(verbose=verbose, num_events=num_events)

    def publish_message(  # noqa: PLR0913
        self,
        link_name: str,
        message: Message[Any],
        qos: int = 0,
        context: Any = None,
        *,
        topic: str = "",
        use_link_topic: bool = False,
    ) -> MQTTMessageInfo:
        if topic and use_link_topic:
            raise ValueError(
                f"Error. Specify at most one of use_link_topic "
                f"({use_link_topic}) and topic ({topic})"
            )
        if not message.Header.Dst:
            message.Header.Dst = self._mqtt_clients.topic_dst(link_name)
        if use_link_topic:
            topic = MQTTTopic.encode(
                envelope_type=message.type_name(),
                src=self.publication_name,
                dst=self._mqtt_clients.topic_dst(link_name),
                message_type=message.message_type(),
            )
        elif not topic:
            topic = message.mqtt_topic()
        payload = self._mqtt_codecs[link_name].encode(message)
        self._logger.message_summary(
            direction="OUT mqtt    ",
            src=message.Header.Src,
            dst=message.Header.Dst,
            topic=topic,
            payload_object=message.Payload,
            message_id=message.Payload.AckMessageID
            if isinstance(message.Payload, Ack)
            else message.Header.MessageId,
        )
        if message.Header.AckRequired:
            self._acks.start_ack_timer(
                link_name, message.Header.MessageId, context=context
            )
        self._message_times.update_send(link_name)
        return self._mqtt_clients.publish(link_name, topic, payload, qos)

    def publish_upstream(
        self, payload: Any, qos: QOS = QOS.AtMostOnce, **message_args: Any
    ) -> MQTTMessageInfo:
        message = Message[Any](
            Src=self.publication_name,
            Dst=self._mqtt_clients.upstream_topic_dst,
            Payload=payload,
            **message_args,
        )
        return self.publish_message(
            self._mqtt_clients.upstream_client, message, qos=qos
        )

    def generate_event(self, event: EventT) -> Result[bool, Exception]:
        path_dbg = 0
        error_count = 0
        warning_count = 0
        self._logger.path(
            "++generate_event %s  f: %d  (p: %d)  (pr: %d  r: %d  c: %d)",
            event.TypeName,
            len(self._in_flight_events),
            self._event_persister.num_pending,
            self._event_persister.num_persists,
            self._event_persister.num_retrieves,
            self._event_persister.num_clears,
        )
        if not event.Src:
            path_dbg |= 0x00000001
            event.Src = self.publication_name
        if (
            isinstance(event, CommEvent)
            and event.Src == self.publication_name
            and self._stats.has_link(event.PeerName)
        ):
            path_dbg |= 0x00000002
            self._stats.link(event.PeerName).comm_event_counts[event.TypeName] += 1
        if isinstance(event, ProblemEvent) and self._logger.path_enabled:
            path_dbg |= 0x00000004
            self._logger.info(f"ProblemEvent <{event.Summary}>\n{event.Details}")  # noqa: G004
        if (
            self._mqtt_clients.upstream_client
            and self._states[self._mqtt_clients.upstream_client].active()
        ):
            path_dbg |= 0x00000008
            if len(self._in_flight_events) >= self._settings.num_inflight_events:
                path_dbg |= 0x00000010
                result = self._event_persister.persist(
                    event.MessageId,
                    event.model_dump_json().encode(PERSISTER_ENCODING),
                )
            else:
                path_dbg |= 0x00000020
                self._in_flight_events[event.MessageId] = event
                result = Ok()
            self.publish_upstream(event, AckRequired=True)
        else:
            path_dbg |= 0x00000040
            result = self._event_persister.persist(
                event.MessageId, event.model_dump_json().encode(PERSISTER_ENCODING)
            )
            match result:
                case Err(problems):
                    error_count = len(problems.errors)
                    warning_count = len(problems.warnings)

        self._logger.path(
            "--generate_event %s  f: %d  (p: %d)  (pr: %d  r: %d  c: %d)  problems: (%d %d)  path:0x%08X",
            event.TypeName,
            len(self._in_flight_events),
            self._event_persister.num_pending,
            self._event_persister.num_persists,
            self._event_persister.num_retrieves,
            self._event_persister.num_clears,
            error_count,
            warning_count,
            path_dbg,
        )
        return result

    def _start_reupload(self) -> None:
        if not self._reuploads.reuploading():
            self._continue_reupload(
                self._reuploads.start_reupload(self._event_persister.pending_ids())
            )

    def _continue_reupload(self, event_ids: list[str]) -> None:
        self._logger.path("++_continue_reupload  %d", len(event_ids))
        path_dbg = 0
        tried_count_dbg = 0
        sent_count_dbg = 0
        continuation_count_dbg = -1

        if event_ids:
            path_dbg |= 0x00000001
            sent_one = False
            # Try to send all requested events. At least send must succeed to
            # continue the reupload, so if all sends fail, get more until
            # one is sent or there are no more reuploads.
            while not sent_one and self._reuploads.reuploading() and event_ids:
                continuation_path_dbg = 0x00000002
                continuation_count_dbg += 1
                next_event_ids = []
                for event_id in event_ids:
                    event_path_dbg = 0x00000004
                    tried_count_dbg += 1
                    problems = Problems()

                    match ret := self._reupload_event(event_id):
                        case Ok():
                            event_path_dbg |= 0x00000008
                            if ret.value:
                                path_dbg |= 0x00000010
                                sent_count_dbg += 1
                                sent_one = True
                            else:
                                event_path_dbg |= 0x00000020
                                problems.add_error(DecodingError(uid=event_id))
                        case Err():
                            event_path_dbg |= 0x00000040
                            problems.add_problems(ret.err())
                    if problems:
                        event_path_dbg |= 0x00000080
                        # There was some error decoding this event.
                        # We generate a new event with information
                        # about decoding failure and delete this event.
                        self.generate_event(
                            problems.problem_event(
                                f"Event decoding error - uid:{event_id}"
                            )
                        )
                        self._event_persister.clear(event_id)
                        if sent_one:
                            event_path_dbg |= 0x00000100
                            self._reuploads.clear_unacked_event(event_id)
                        else:
                            event_path_dbg |= 0x00000200
                            next_event_ids.extend(
                                self._reuploads.process_ack_for_reupload(event_id)
                            )
                    self._logger.path("  1 event path:0x%08X", event_path_dbg)
                    continuation_path_dbg |= event_path_dbg
                self._logger.path("  1 continuation path:0x%08X", continuation_path_dbg)
                event_ids = next_event_ids
                path_dbg |= continuation_path_dbg
        self._logger.path(
            "--_continue_reupload  path:0x%08X  sent:%d  tried:%d  continuations:%d",
            path_dbg,
            sent_count_dbg,
            tried_count_dbg,
            continuation_count_dbg,
        )

    def _reupload_event(self, event_id: str) -> Result[bool, Problems]:
        """Load event for event_id from storage, decoded to JSON and send it.

        Return either Ok(True) or Err(Problems(list of decoding errors)).

        Send errors handled either by exception, which will propagate up, or
        by ack timeout.
        """
        self._logger.path("++_reupload_event  %s", event_id)
        path_dbg = 0
        problems = Problems()
        match self._event_persister.retrieve(event_id):
            case Ok(event_bytes):
                path_dbg |= 0x00000001
                if event_bytes is None:
                    path_dbg |= 0x00000002
                    problems.add_error(
                        UIDMissingWarning("reupload_events", uid=event_id)
                    )
                elif len(event_bytes) == 0:
                    path_dbg |= 0x00000004
                    problems.add_error(
                        FileEmptyWarning("reupload_events", uid=event_id)
                    )
                else:
                    path_dbg |= 0x00000008
                    try:
                        event_str = event_bytes.decode(encoding=self.PERSISTER_ENCODING)
                    except Exception as e:  # noqa: BLE001
                        path_dbg |= 0x00000010
                        problems.add_error(e).add_error(
                            ByteDecodingError("reupload_events", uid=event_id)
                        )
                    else:
                        path_dbg |= 0x00000020
                        try:
                            event = json.loads(event_str)
                        except Exception as e:  # noqa: BLE001
                            path_dbg |= 0x00000040
                            problems.add_error(e).add_error(
                                JSONDecodingError(
                                    f"reupload_events - raw json:\n<\n{event_str}\n>",
                                    uid=event_id,
                                )
                            )
                        else:
                            path_dbg |= 0x00000080
                            self.publish_upstream(event, AckRequired=True)
                            self._logger.path(
                                "--_reupload_event:1  path:0x%08X", path_dbg
                            )
                            return Ok(value=True)
            case Err(error):
                path_dbg |= 0x00000100
                problems.add_problems(error)
        self._logger.path("--_reupload_event:0  path:0x%08X", path_dbg)
        return Err(problems)

    def start(
        self, loop: asyncio.AbstractEventLoop, async_queue: asyncio.Queue[Any]
    ) -> None:
        self._logger.path("++LinkManager.start")
        if self.upstream_client:
            self._reuploads.stats = self._stats.link(self.upstream_client)
        self._mqtt_clients.start(loop, async_queue)
        self.generate_event(StartupEvent())
        self._states.start_all()
        self._logger.path("--LinkManager.start")

    def stop(self) -> Result[bool, Problems]:
        problems: Optional[Problems] = None
        for link_name in self._states.link_names():
            ret = self._states.stop(link_name)
            match ret := self._states.stop(link_name):
                case Err():
                    if problems is None:
                        problems = Problems(errors=[ret.err()])
                    else:
                        problems.errors.append(ret.err())
        self._mqtt_clients.stop()
        if problems is None:
            return Ok(value=True)
        return Err(problems)

    def process_mqtt_connected(
        self, message: Message[MQTTConnectPayload]
    ) -> Result[Transition, InvalidCommStateInput]:
        state_result = self._states.process_mqtt_connected(message)
        if state_result.is_ok():
            self._logger.comm_event(str(state_result.value))
        self.generate_event(MQTTConnectEvent(PeerName=message.Payload.client_name))
        self._mqtt_clients.subscribe_all(message.Payload.client_name)
        return state_result

    def process_mqtt_disconnected(
        self, message: Message[MQTTDisconnectPayload]
    ) -> Result[LinkManagerTransition, InvalidCommStateInput]:
        state_result = self._states.process_mqtt_disconnected(message)
        result: Result[LinkManagerTransition, InvalidCommStateInput]
        match state_result:
            case Ok():
                # noinspection PyTypeChecker
                result = Ok(LinkManagerTransition(**(asdict(state_result.value))))
                self.generate_event(
                    MQTTDisconnectEvent(PeerName=message.Payload.client_name)
                )
                self._logger.comm_event(str(result.value))
                if result.value.recv_deactivated() or result.value.send_deactivated():
                    result.value.canceled_acks = self._acks.cancel_ack_timers(
                        message.Payload.client_name
                    )
                    self._reuploads.clear()
                    self.flush_in_flight_events()
            case _:
                result = state_result
        return result

    def flush_in_flight_events(self) -> None:
        for event_id, event in self._in_flight_events.items():
            self._event_persister.persist(
                event_id, event.model_dump_json().encode(PERSISTER_ENCODING)
            )
        self._in_flight_events.clear()

    def process_mqtt_connect_fail(
        self, message: Message[MQTTConnectFailPayload]
    ) -> Result[Transition, InvalidCommStateInput]:
        return self._states.process_mqtt_connect_fail(message)

    def process_mqtt_message(
        self, message: Message[MQTTReceiptPayload]
    ) -> Result[Transition, InvalidCommStateInput]:
        match result := self._states.process_mqtt_message(message):
            case Ok():
                self.update_recv_time(message.Payload.client_name)
                if result.value:
                    self._logger.comm_event(str(result.value))
                if result.value.recv_activated():
                    self._recv_activated(result.value)
        return result

    def process_ack_timeout(
        self, wait_info: AckWaitInfo
    ) -> Result[LinkManagerTransition, Exception]:
        self._logger.path(
            "++LinkManager.process_ack_timeout  <%s>  <%s>",
            wait_info.link_name,
            wait_info.message_id,
        )
        path_dbg = 0
        self._stats.link(wait_info.link_name).timeouts += 1
        result: Result[LinkManagerTransition, Exception]
        match state_result := self._states.process_ack_timeout(wait_info.link_name):
            case Ok():
                path_dbg |= 0x00000001
                transition = LinkManagerTransition(
                    canceled_acks=[wait_info], **(asdict(state_result.value))
                )
                if transition.deactivated():
                    path_dbg |= 0x00000002
                    self.flush_in_flight_events()
                    self._reuploads.clear()
                    self.generate_event(
                        ResponseTimeoutEvent(PeerName=transition.link_name)
                    )
                    self._logger.comm_event(str(transition))
                    transition.canceled_acks.extend(
                        self._acks.cancel_ack_timers(wait_info.link_name)
                    )
                result = Ok(transition)
            case _:
                result = Err(state_result.err())
        self._logger.path("--LinkManager.process_ack_timeout path:0x%08X", path_dbg)
        return result

    def process_ack(self, link_name: str, message_id: str) -> None:
        self._logger.path("++LinkManager.process_ack  <%s>  %s", link_name, message_id)
        path_dbg = 0
        wait_info = self._acks.cancel_ack_timer(link_name, message_id)
        if wait_info is not None:
            path_dbg |= 0x00000001
            if message_id in self._in_flight_events:
                path_dbg |= 0x00000002
                self._in_flight_events.pop(message_id)
            elif message_id in self._event_persister:
                path_dbg |= 0x00000004
                self._event_persister.clear(message_id)
                if self._reuploads.reuploading() and link_name == self.upstream_client:
                    path_dbg |= 0x00000008
                    self._continue_reupload(
                        self._reuploads.process_ack_for_reupload(message_id)
                    )
                    if not self._reuploads.reuploading():
                        path_dbg |= 0x00000010
                        self._logger.info("reupload complete.")
        self._logger.path("--LinkManager.process_ack path:0x%08X", path_dbg)

    def send_ack(self, link_name: str, message: Message[Any]) -> None:
        if message.Header.MessageId:
            self.publish_message(
                link_name,
                Message(
                    Src=self.publication_name,
                    Payload=Ack(AckMessageID=message.Header.MessageId),
                ),
            )

    def start_ping_tasks(self) -> list[asyncio.Task[Any]]:
        return [
            asyncio.create_task(
                self.send_ping(link_name), name=f"send_ping<{link_name}>"
            )
            for link_name in self.link_names()
        ]

    async def send_ping(self, link_name: str) -> None:
        while not self._states.stopped(link_name):
            message_times = self._message_times.get_copy(link_name)
            link_state = self._states[link_name]
            if (
                message_times.time_to_send_ping(self._settings.mqtt_link_poll_seconds)
                and link_state.active_for_send()
            ):
                self.publish_message(link_name, PingMessage(Src=self.publication_name))
            await asyncio.sleep(
                message_times.seconds_until_next_ping(
                    self._settings.mqtt_link_poll_seconds
                )
            )

    def update_recv_time(self, link_name: str) -> None:
        self._message_times.update_recv(link_name)

    def _recv_activated(self, transition: Transition) -> None:
        self._logger.path("++LinkManager._recv_activated.<%s>", transition)
        path_dbg = 0
        if transition.link_name == self.upstream_client:
            path_dbg |= 0x00000001
            self._start_reupload()
        self.generate_event(PeerActiveEvent(PeerName=transition.link_name))
        self._logger.path(
            "--LinkManager._recv_activated.<%s>  path:0x%08X", transition, path_dbg
        )

    def process_mqtt_suback(
        self, message: Message[MQTTSubackPayload]
    ) -> Result[Transition, InvalidCommStateInput]:
        self._logger.path(
            "++LinkManager.process_mqtt_suback client:%s", message.Payload.client_name
        )
        path_dbg = 0
        state_result = self._states.process_mqtt_suback(
            message.Payload.client_name,
            self._mqtt_clients.handle_suback(message.Payload),
        )
        if isinstance(state_result, Ok):
            path_dbg |= 0x00000001
            if state_result.value:
                path_dbg |= 0x00000002
                self._logger.comm_event(str(state_result.value))
            if state_result.value.send_activated():
                path_dbg |= 0x00000004
                self.generate_event(
                    MQTTFullySubscribedEvent(PeerName=message.Payload.client_name)
                )
                self.publish_message(
                    message.Payload.client_name,
                    PingMessage(Src=self.publication_name),
                )
            if state_result.value.recv_activated():
                path_dbg |= 0x00000008
                self._recv_activated(state_result.value)
        self._logger.path(
            "--LinkManager.process_mqtt_suback:%d  path:0x%08X",
            state_result.is_ok(),
            path_dbg,
        )
        return state_result



"""Generate or copy test certificates for MQTT using TLS."""

# ruff: noqa: T201, S101

import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Optional

import rich
from gwcert import DEFAULT_CA_DIR
from pydantic import BaseModel
from pydantic_settings import BaseSettings

from gwproactor.config.mqtt import MQTTClient
from gwproactor.config.paths import TLSPaths

TEST_CA_CERTIFICATE_PATH_VAR = "GWPROACTOR_TEST_CA_CERT_PATH"
"""Environment variable containing the path to the CA certificate used by the tests."""

TEST_CA_PRIVATE_KEY_VAR = "GWPROACTOR_TEST_CA_KEY_PATH"
"""Environment variable containing the the path to the CA private key used by the tests."""

TEST_CERTIFICATE_CACHE_VAR = "GWPROACTOR_TEST_CERTIFICATE_CACHE"
"""Environment variable containing the path to the cache of TLS certificates used by the tests."""

TEST_CA_CERTIFICATE_PATH = DEFAULT_CA_DIR / "ca.crt"
TEST_CA_PRIVATE_KEY_PATH = DEFAULT_CA_DIR / "private" / "ca_key.pem"


def _key_path(tag: str, var_name: str, default: Path) -> Path:
    """Return the path to a specified key, specified by environment variable or default. Raise an
    error if that a file that path does not exist."""
    env_val = os.getenv(var_name)
    if env_val:
        key_path = Path(env_val)
        if not key_path.exists():
            raise ValueError(
                f"ERROR. Key <{tag}> does not exist at path: {key_path}.\n"
                f"  Key path specified via environment variable {var_name}"
            )
    else:
        key_path = default
        if not key_path.exists():
            raise ValueError(
                f"ERROR. Key <{tag}> does not exist at path: {key_path}.\n"
                f"  Key path is the default path for <{tag}> because environment variable {var_name} is empty or missing."
            )
    return key_path


def test_ca_certificate_path() -> Path:
    """Return the path to the CA certificate used in testing, specified by environment variable or default. Raise an
    error if that a file that path does not exist.
    """
    return _key_path(
        "CA certificate", TEST_CA_CERTIFICATE_PATH_VAR, TEST_CA_CERTIFICATE_PATH
    )


def test_ca_private_key_path() -> Path:
    """Return the path to the CA private key used in testing, specified by environment variable or default. Raise an
    error if that a file that path does not exist.
    """
    return _key_path(
        "CA private key", TEST_CA_PRIVATE_KEY_VAR, TEST_CA_PRIVATE_KEY_PATH
    )


def _copy_keys(test_cert_dir: Path, dst_paths: TLSPaths) -> None:
    ca_cert_path = test_ca_certificate_path()
    ca_private_key_path = test_ca_private_key_path()
    if (
        dst_paths.cert_path is None
        or dst_paths.private_key_path is None
        or dst_paths.ca_cert_path is None
    ):
        raise ValueError(
            f"ERROR. _copy_keys dst_paths ({dst_paths}) must have non-None paths"
        )
    src_paths = TLSPaths(
        ca_cert_path=ca_cert_path,
        cert_path=test_cert_dir / Path(dst_paths.cert_path).name,
        private_key_path=test_cert_dir
        / "private"
        / Path(dst_paths.private_key_path).name,
    )
    assert isinstance(src_paths.ca_cert_path, Path)
    assert isinstance(src_paths.cert_path, Path)
    assert isinstance(src_paths.private_key_path, Path)
    if not src_paths.cert_path.exists() or not src_paths.private_key_path.exists():
        print(
            f"One or more TLS test keys at {test_cert_dir} does not exist. Recreating."
        )
        print(
            f"  Cert path        exists: {src_paths.cert_path.exists()!s:5s}  {src_paths.cert_path}"
        )
        print(
            f"  Private key path exists: {src_paths.private_key_path.exists()!s:5s}  {src_paths.private_key_path}"
        )
        gwcert_command = [
            "gwcert",
            "key",
            "add",
            "--ca-certificate-path",
            str(ca_cert_path),
            "--ca-private-key-path",
            str(ca_private_key_path),
            "--force",
            str(src_paths.cert_path),
        ]
        print(f"Generating keys with command:\n  {' '.join(gwcert_command)}")
        result = subprocess.run(gwcert_command, capture_output=True, check=True)  # noqa: S603
        print(result.stdout.decode("utf-8"))
    dst_paths.mkdirs()
    shutil.copy2(src_paths.ca_cert_path, dst_paths.ca_cert_path)
    shutil.copy2(src_paths.cert_path, dst_paths.cert_path)
    shutil.copy2(src_paths.private_key_path, dst_paths.private_key_path)


def mqtt_client_fields(model: BaseModel | BaseSettings) -> list[tuple[str, MQTTClient]]:
    clients: list[tuple[str, MQTTClient]] = [
        (field_name, getattr(model, field_name))
        for field_name in model.__pydantic_fields__
        if isinstance(getattr(model, field_name), MQTTClient)
    ]
    clients.extend(
        [
            (field_name, field_value)
            for field_name, field_value in getattr(
                model, "mqtt", getattr(model, "mqtt_brokers", {})
            ).items()
            if isinstance(field_value, MQTTClient)
        ]
    )
    return clients


def uses_tls(model: BaseModel | BaseSettings) -> bool:
    """Check whether any MQTTClient in the model have MQTTClient.tls.use_tls == True."""
    return any(
        client_field[1].tls.use_tls for client_field in mqtt_client_fields(model)
    )


def copy_keys(
    model_tag: str,
    model: BaseModel | BaseSettings,
    test_cert_cache_dir: Optional[Path] = None,
) -> None:
    """Copy keys referenced in a model from the cache to locations specified by the model."""
    if test_cert_cache_dir is None:
        test_cert_cache_dir = test_certificate_cache_dir()
    test_cert_cache = Path(test_cert_cache_dir)
    for field_name, client in mqtt_client_fields(model):
        _copy_keys(test_cert_cache / model_tag / field_name, client.tls.paths)


def set_test_certificate_cache_dir(certificate_cache_dir: Path | str) -> None:
    """Set the GWPROACTOR_TEST_CERTIFICATE_CACHE environment variable.

    May be called in conftest.py as:

        set_test_certificate_cache(Path(__file__).parent / '.certificate_cache')

    """
    found_cache_dir = os.getenv(TEST_CERTIFICATE_CACHE_VAR, "")
    if not found_cache_dir:
        os.environ[TEST_CERTIFICATE_CACHE_VAR] = str(
            Path(certificate_cache_dir).absolute()
        )


def test_certificate_cache_dir() -> Path:
    """Get certificate cache directory, or raise an exception if it is not set."""
    found_cache_dir = os.getenv(TEST_CERTIFICATE_CACHE_VAR, "")
    if not found_cache_dir:
        raise ValueError(
            f"ERROR. Environment variable {TEST_CERTIFICATE_CACHE_VAR} is unset or empty.\n"
            f"  To test with TLS, either set {TEST_CERTIFICATE_CACHE_VAR} explicitly, or "
            "set in base conftest.py with, e.g:\n\n"
            "  set_test_certificate_cache(Path(__file__).parent / '.certificate_cache')"
        )
    return Path(found_cache_dir)


def generate_dummy_certs(settings: BaseModel, *, dry_run: bool = False) -> None:
    """Generate certs for dummy proactors."""
    console = rich.console.Console()
    ca_cert_path = test_ca_certificate_path()
    ca_private_key_path = test_ca_private_key_path()
    commands_ = [
        [
            "gwcert",
            "key",
            "add",
            "--ca-certificate-path",
            str(ca_cert_path),
            "--ca-private-key-path",
            str(ca_private_key_path),
            "--force",
            str(client_config.tls.paths.cert_path),
        ]
        for _, client_config in mqtt_client_fields(settings)
    ]
    if dry_run:
        console.print(f"Showing key generation commands for {type(settings)}")
        for command in commands_:
            console.print(f"  {' '.join(command)}", soft_wrap=True)
        console.print("\n\n\n")
    else:
        console.print(f"Generating keys with for {type(settings)}")
        for command in commands_:
            console.print("Generating keys with command:")
            console.print(f"  {' '.join(command)}", soft_wrap=True)
            try:
                result = subprocess.run(command, capture_output=True, check=True)  # noqa: S603
                print(result.stdout.decode("utf-8"))  # noqa
            except subprocess.CalledProcessError as e:
                print(e.stderr.decode("utf-8"))  # noqa
                sys.exit(e.returncode)
        console.print("\n")



import asyncio
import contextlib
import logging
import threading
import time
from typing import Any, Coroutine, Optional, Sequence

from gwproto import Message
from result import Result

from gwproactor.message import KnownNames, PatInternalWatchdogMessage, ShutdownMessage
from gwproactor.proactor_interface import (
    INVALID_IO_TASK_HANDLE,
    AppInterface,
    Communicator,
    IOLoopInterface,
    MonitoredName,
)
from gwproactor.problems import Problems
from gwproactor.sync_thread import SyncAsyncInteractionThread


class IOLoop(Communicator, IOLoopInterface):
    _io_loop: asyncio.AbstractEventLoop
    _io_thread: Optional[threading.Thread] = None
    _tasks: dict[int, Optional[asyncio.Task[Any]]]
    _task2id: dict[asyncio.Task[Any], int]
    _lock: threading.RLock
    _stop_requested: bool = False
    _next_id = INVALID_IO_TASK_HANDLE + 1
    _pat_timeout: float = SyncAsyncInteractionThread.PAT_TIMEOUT
    _last_pat_time: float = 0.0
    logger: logging.Logger

    def __init__(self, services: AppInterface) -> None:
        super().__init__(KnownNames.io_loop_manager.value, services)
        self.logger = services.logger.io_loop_logger
        self._lock = threading.RLock()
        self._tasks = {}
        self._task2id = {}
        self._io_loop = asyncio.new_event_loop()

    def add_io_coroutine(self, coro: Coroutine[Any, Any, Any], name: str = "") -> int:
        with self._lock:
            if self._stop_requested:
                return INVALID_IO_TASK_HANDLE
            task_id = self._next_id
            self._next_id += 1
            self._tasks[task_id] = None
        self._io_loop.call_soon_threadsafe(self._add_task, coro, name, task_id)
        return task_id

    def cancel_io_routine(self, handle: int) -> None:
        with self._lock:
            task = self._tasks.pop(handle, None)
        if task is not None:
            self._io_loop.call_soon_threadsafe(task.cancel, "IOLoop.cancel_io_routine")

    def _add_task(
        self, coro: Coroutine[Any, Any, Any], name: str, task_id: int
    ) -> None:
        if not name:
            name = coro.__name__
        task = self._io_loop.create_task(
            coro,
            name=name,
        )
        with self._lock:
            self._tasks[task_id] = task
            self._task2id[task] = task_id

    def start(self) -> None:
        self._io_thread = threading.Thread(
            name=self.name, target=self._thread_run, daemon=True
        )
        self._io_thread.start()

    def _started_tasks(self) -> list[asyncio.Task[Any]]:
        with self._lock:
            return [task for task in self._tasks.values() if task is not None]

    def stop(self) -> None:
        with self._lock:
            self._stop_requested = True
            tasks = self._started_tasks()
        for task in tasks:
            self._io_loop.call_soon_threadsafe(task.cancel, "IOLoop.stop")

    async def join(self) -> None:
        pass
        # this often generates errors, although tests pass:
        # await async_polling_thread_join(self._io_thread)  # noqa: ERA001

    def _thread_run(self) -> None:
        try:
            self._io_loop.run_until_complete(self._async_run())
        except Exception as e:  # noqa: BLE001
            try:
                summary = f"Unexpected IOLoop exception <{e}>"
            except:  # noqa: E722
                summary = "Unexpected IOLoop exception"
            with contextlib.suppress(Exception):
                self._services.send_threadsafe(
                    Message(
                        Payload=Problems(errors=[e]).problem_event(
                            summary=summary, src=self.name
                        )
                    )
                )
            with contextlib.suppress(Exception):
                self._services.send_threadsafe(
                    ShutdownMessage(
                        Src=self.name,
                        Reason=summary,
                    )
                )
        finally:
            self._io_loop.stop()

    async def _async_run(self) -> None:  # noqa: C901, PLR0912
        self.logger.info("++IOLoop._async_run")
        path_dbg = 0
        try:  # noqa: PLR1702
            while not self._stop_requested:
                self.logger.debug("++IOLoop._async_run loop")
                path_dbg = 0
                with self._lock:
                    tasks = self._started_tasks()
                if not self._stop_requested and not tasks:
                    path_dbg |= 0x00000001
                    try:
                        await asyncio.sleep(1)
                    except asyncio.CancelledError:
                        path_dbg |= 0x00000002
                    except GeneratorExit:
                        path_dbg |= 0x00000004
                else:
                    path_dbg |= 0x00000008
                    done, _ = await asyncio.wait(
                        tasks, timeout=1.0, return_when="FIRST_COMPLETED"
                    )
                    if self._stop_requested:
                        path_dbg |= 0x00000010  # type: ignore[unreachable]
                        break
                    with self._lock:
                        for task in done:
                            task_id = self._task2id.pop(task)
                            # Suppress false positive warning from PyCharm
                            # (https://youtrack.jetbrains.com/issue/PY-51101/False-positive-Coroutine-...-is-not-awaited-for-a-non-coroutine-call)
                            # noinspection PyAsyncCall
                            self._tasks.pop(task_id)
                    errors = []
                    for task in done:
                        path_dbg |= 0x00000020
                        if not task.cancelled():
                            path_dbg |= 0x00000040
                            try:
                                exception = task.exception()
                                if exception is not None:
                                    path_dbg |= 0x00000080
                                    errors.append(exception)
                            except Exception as retrieve_exception:  # noqa: BLE001
                                path_dbg |= 0x00000100
                                errors.append(retrieve_exception)
                    if errors:
                        path_dbg |= 0x00000200
                        raise Problems(
                            f"IOLoop caught {len(errors)}.",
                            errors=errors,
                        )
                if self.time_to_pat():
                    path_dbg |= 0x00000400
                    self.pat_watchdog()
                self.logger.debug("--IOLoop._async_run loop  path:0X%08X", path_dbg)
        finally:
            self._io_loop.stop()
            self.logger.info("--IOLoop._async_run  path:0X%08X", path_dbg)

    def time_to_pat(self) -> bool:
        return time.time() >= (self._last_pat_time + (self._pat_timeout / 2))

    def pat_watchdog(self) -> None:
        if not self._stop_requested:
            self._last_pat_time = time.time()
            self._services.send_threadsafe(PatInternalWatchdogMessage(src=self.name))

    def process_message(self, message: Message[Any]) -> Result[bool, Exception]:  # noqa: ARG002
        raise ValueError("IOLoop does not currently process any messages")

    def _send(self, message: Message[Any]) -> None:
        """Ensure this routine from base class does not make unsafe call."""
        self._services.send_threadsafe(message)

    @property
    def monitored_names(self) -> Sequence[MonitoredName]:
        return [MonitoredName(self.name, self._pat_timeout)]



from gwproto import HardwareLayout, Message, MQTTTopic

from gwproactor import App, AppSettings, Proactor
from gwproactor.config import MQTTClient
from gwproactor.config.links import LinkSettings
from gwproactor.config.proactor_config import ProactorName
from gwproactor.links import QOS
from gwproactor.persister import PersisterInterface, TimedRollingFilePersister
from gwproactor_test.dummies import DUMMY_CHILD_NAME, DUMMY_PARENT_NAME


class DummyChildSettings(AppSettings):
    parent: MQTTClient = MQTTClient()


class DummyChildApp(App):
    PARENT_MQTT: str = DUMMY_PARENT_NAME

    @classmethod
    def app_settings_type(cls) -> type[DummyChildSettings]:
        return DummyChildSettings

    @classmethod
    def paths_name(cls) -> str:
        return DUMMY_CHILD_NAME

    def _get_name(self, layout: HardwareLayout) -> ProactorName:
        return ProactorName(
            long_name=layout.scada_g_node_alias,
            short_name="s",
        )

    def _get_link_settings(
        self,
        name: ProactorName,  # noqa: ARG002
        layout: HardwareLayout,
        brokers: dict[str, MQTTClient],  # noqa: ARG002
    ) -> dict[str, LinkSettings]:
        return {
            self.PARENT_MQTT: LinkSettings(
                broker_name=self.PARENT_MQTT,
                peer_long_name=layout.atn_g_node_alias,
                peer_short_name="a",
                upstream=True,
            )
        }

    @classmethod
    def _make_persister(cls, settings: AppSettings) -> PersisterInterface:
        return TimedRollingFilePersister(settings.paths.event_dir)

    def _connect_links(self, proactor: Proactor) -> None:
        super()._connect_links(proactor)
        for topic in [
            MQTTTopic.encode_subscription(Message.type_name(), "1", "a"),
            MQTTTopic.encode_subscription(Message.type_name(), "2", "b"),
        ]:
            proactor.links.subscribe(self.PARENT_MQTT, topic, QOS.AtMostOnce)



# ruff: noqa: T201
import abc
import cProfile
import pstats
import shutil
import time
import uuid
from enum import Enum
from pathlib import Path
from pstats import SortKey
from typing import NamedTuple, Optional

import rich
import typer
from pydantic import BaseModel

from gwproactor.config import Paths
from gwproactor.persister import (
    PersisterInterface,
    TimedRollingFilePersister,
)

MEASUREMENT_FILE = "persister-measurements.json"

app = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    rich_markup_mode="rich",
    help="Measure persister performance",
)


class PersisterType(str, Enum):
    json = "json"
    all = "all"


persister_classes: dict[PersisterType, type] = {
    PersisterType.json: TimedRollingFilePersister,
}


def paths_name(persister_type: PersisterType) -> str:
    return f"persister-test-{persister_type.value}"


class MeasurementParams(NamedTuple):
    persister_type: str
    func_name: str
    n: int
    measurements: int
    measure_size: int

    def __str__(self) -> str:
        return f"{self.persister_type}-{self.func_name}-n{self.n}-m{self.measurements}-ms{self.measure_size}"


class FuncMeasurements(BaseModel):
    params: MeasurementParams
    i: list[int] = []
    single_times: list[float] = []
    step_times: list[float] = []
    total_time: float = 0


class Measurements(BaseModel):
    funcs: dict[str, FuncMeasurements] = {}


class MeasureCommand(abc.ABC):
    paths: Paths
    measurements_dir: Path
    measurement_file: Path
    stats_file: Path
    persister: PersisterInterface
    params: MeasurementParams

    def __init__(
        self,
        persister_type: PersisterType,
        n: Optional[int] = None,
        measurements: int = 10,
        *,
        do_reindex: bool = True,
        persister: Optional[PersisterInterface] = None,
        paths: Optional[Paths] = None,
    ) -> None:
        if persister_type not in persister_classes:
            raise ValueError(f"Unexpected persister type {persister_type}")
        if paths is None:
            self.paths = Paths(name=paths_name(persister_type))
        else:
            self.paths = paths
        self.paths.mkdirs()
        self.measurements_dir = Path(self.paths.data_dir) / "measurements"
        self.measurements_dir.mkdir(exist_ok=True, parents=True)
        self.measurement_file = self.measurements_dir / MEASUREMENT_FILE
        if persister is None:
            self.persister = persister_classes[persister_type](self.paths.event_dir)
            if do_reindex:
                self.persister.reindex()
        else:
            self.persister = persister
        if n is None:
            n = self.persister.num_pending
        measurements = min(measurements, n)
        self.params = MeasurementParams(
            persister_type=persister_type.value,
            func_name=self.get_name(),
            n=n,
            measurements=measurements,
            measure_size=n // measurements if n else 0,
        )
        self.stats_file = self.measurements_dir / f"{self.params}.prf"

    def measure(
        self,
        *,
        profile: bool = False,
    ) -> FuncMeasurements:
        data = FuncMeasurements(params=self.params)
        step_start_time = time.time()
        call_start_time: float = step_start_time
        prf: Optional[cProfile.Profile]
        start_size_str = self.size_str()
        if profile:
            prf = cProfile.Profile()
            prf.enable()
        else:
            prf = None
        event_ids = self.get_event_ids()
        total_start_time = time.time()
        for i in range(data.params.n):
            measure_now = (i + 1) % data.params.measure_size == 0
            if measure_now:
                call_start_time = time.time()
            self.func_to_measure(event_ids[i])
            if measure_now:
                end_time = time.time()
                data.i.append(i)
                data.single_times.append(end_time - call_start_time)
                data.step_times.append(end_time - step_start_time)
                step_start_time = end_time
        data.total_time = time.time() - total_start_time
        if prf is not None:
            prf.disable()
            prf.create_stats()
            prf.dump_stats(self.stats_file)
            sortby = SortKey.CUMULATIVE
            ps = pstats.Stats(prf).sort_stats(sortby)
            rich.print(f"Profiling for {data.params}:")
            ps.print_stats(20)
        else:
            rich.print(f"Measurements for {data.params}:")
            rich.print(data)
            if self.measurement_file.exists():
                with self.measurement_file.open("r") as f:
                    read = f.read()
                if read:
                    recorded = Measurements.model_validate_json(read)
                else:
                    recorded = Measurements()
            else:
                recorded = Measurements()
            param_str = str(data.params)
            if param_str in recorded.funcs:
                existing = recorded.funcs[param_str]
                existing.i.extend(data.i)
                existing.single_times.extend(data.single_times)
                existing.step_times.extend(data.step_times)
            else:
                recorded.funcs[param_str] = data
            with self.measurement_file.open("w") as f:
                f.write(recorded.model_dump_json(indent=2))
        rich.print(f"Size before: {start_size_str}")
        rich.print(f"Size after:  {self.size_str()}")
        return data

    @abc.abstractmethod
    def func_to_measure(self, event_id: str) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def get_event_ids(self) -> list[str]:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def get_name(cls) -> str:
        raise NotImplementedError

    def size_str(self) -> str:
        curr_bytes = self.persister.curr_bytes
        curr_KB = curr_bytes / 1024
        curr_KB_str = f"{curr_KB:12.1f}" if curr_KB < 1 else f"{int(curr_KB):14d}"
        curr_MB = curr_KB / 1024
        return f"{curr_bytes:14d} bytes    {curr_KB_str} KB    {curr_MB:6.1f} MB"

    def __str__(self) -> str:
        return (
            f"Commands: {self.__class__.__name__}\n"
            f"  Persister: {self.persister.__class__.__name__}\n"
            f"  Measure params: {self.params}\n"
            f"  Data dir:\n    {self.paths.data_dir}\n"
            f"  Measurement file:\n    {self.measurement_file}\n"
            f"  Stats file:\n    {self.stats_file}\n"
        )


class MeasurePersist(MeasureCommand):
    record: bytes

    def __init__(
        self,
        persister_type: PersisterType,
        n: int = 10000,
        measurements: int = 10,
        rec_size: int = 10 * 1024,
        *,
        persister: Optional[PersisterInterface] = None,
        paths: Optional[Paths] = None,
    ) -> None:
        self.record = ("#" * rec_size).encode()
        super().__init__(
            persister_type=persister_type,
            n=n,
            measurements=measurements,
            paths=paths,
            persister=persister,
        )

    def func_to_measure(self, event_id: str) -> None:
        self.persister.persist(event_id, self.record)

    def get_event_ids(self) -> list[str]:
        return [str(uuid.uuid4()) for _ in range(self.params.n)]

    @classmethod
    def get_name(cls) -> str:
        return "persist"


class MeasureRetrieve(MeasureCommand):
    def func_to_measure(self, event_id: str) -> None:
        self.persister.retrieve(event_id)

    def get_event_ids(self) -> list[str]:
        return self.persister.pending_ids()

    @classmethod
    def get_name(cls) -> str:
        return "retrieve"


class MeasureClear(MeasureCommand):
    def func_to_measure(self, event_id: str) -> None:
        self.persister.clear(event_id)

    def get_event_ids(self) -> list[str]:
        return self.persister.pending_ids()

    @classmethod
    def get_name(cls) -> str:
        return "clear"


class MeasureReindex(MeasureCommand):
    def __init__(
        self,
        persister_type: PersisterType,
        *,
        persister: Optional[PersisterInterface] = None,
        paths: Optional[Paths] = None,
    ) -> None:
        super().__init__(
            persister_type=persister_type,
            do_reindex=False,
            n=1,
            measurements=1,
            paths=paths,
            persister=persister,
        )

    def func_to_measure(self, _: str) -> None:
        self.persister.reindex()

    def get_event_ids(self) -> list[str]:
        return [""]

    @classmethod
    def get_name(cls) -> str:
        return "reindex"


@app.command()
def persist(
    persister_type: PersisterType,
    n: int = 10000,
    rec_size: int = 10 * 1024,
    measurements: int = 10,
    profile: bool = False,
    dry_run: bool = False,
) -> None:
    """Measure timing of PersisterInterface.persist()"""
    command = MeasurePersist(
        persister_type=persister_type, n=n, measurements=measurements, rec_size=rec_size
    )
    rich.print(command)
    if not dry_run:
        command.measure(profile=profile)


@app.command()
def reindex(
    persister_type: PersisterType, profile: bool = False, dry_run: bool = False
) -> None:
    """Measure timing of PersisterInterface.reindex()"""
    command = MeasureReindex(persister_type=persister_type)
    rich.print(command)
    if not dry_run:
        command.measure(profile=profile)


@app.command()
def retrieve(
    persister_type: PersisterType,
    n: Optional[int] = None,
    measurements: int = 10,
    profile: bool = False,
    dry_run: bool = False,
) -> None:
    """Measure timing of PersisterInterface.retrieve()"""
    command = MeasureRetrieve(
        persister_type=persister_type, n=n, measurements=measurements
    )
    rich.print(command)
    if not dry_run:
        command.measure(profile=profile)


@app.command()
def clear(
    persister_type: PersisterType,
    n: Optional[int] = None,
    measurements: int = 10,
    profile: bool = False,
    dry_run: bool = False,
) -> None:
    """Measure timing of PersisterInterface.retrieve()"""
    command = MeasureClear(
        persister_type=persister_type, n=n, measurements=measurements
    )
    rich.print(command)
    if not dry_run:
        command.measure(profile=profile)


@app.command()
def size(persister_type: PersisterType, dry_run: bool = False) -> None:
    """Print the storage size used by the requested persister."""
    command = MeasureReindex(persister_type=persister_type)
    if not dry_run:
        command.persister.reindex()
        rich.print(command.size_str())


@app.command()
def cleanup(
    persister_type: PersisterType,
    dry_run: bool = False,
    measurements: bool = False,
) -> None:
    """Delete event directories and, optionally, measurements."""
    persister_types = (
        persister_classes.keys()
        if persister_type == PersisterType.all
        else [persister_type]
    )
    for persister_type_ in persister_types:
        rich.print(f"Cleanup for persister {persister_type_.value}")
        paths = Paths(name=paths_name(persister_type_))
        event_dir = Path(paths.event_dir)
        rich.print(f"Removing {event_dir}")
        if not dry_run and event_dir.exists():
            shutil.rmtree(str(event_dir))
        if measurements:
            measurement_dir = Path(paths.data_dir) / "measurements"
            rich.print(f"Removing {measurement_dir}")
            if not dry_run and measurement_dir.exists():
                shutil.rmtree(str(measurement_dir))


def delimit_str(tag: str, markup: str) -> str:
    s = "\n"
    s += f"[{markup}]" + "#" * 80 + "\n"
    s += f"#[/] {tag}\n"
    s += f"[{markup}]" + "#" * 80
    return s


def delimit(tag: str, markup: str) -> None:
    rich.print(delimit_str(tag, markup))


class TotalTime(NamedTuple):
    func_name: str
    time: float


def _persister_suite(
    persister_type: PersisterType,
    n: int = 10000,
    profile: bool = False,
    clean_measurements: bool = False,
    dry_run: bool = False,
) -> list[TotalTime]:
    profile_str = "with profiling" if profile else "without profiling"
    delimit_tag = f"{persister_type.value}  {profile_str}"
    delimit(delimit_tag, "orange3")
    cleanup(
        persister_type=persister_type, measurements=clean_measurements, dry_run=dry_run
    )
    times: list[TotalTime] = []
    measure_persist = MeasurePersist(persister_type=persister_type, n=n)
    paths = measure_persist.paths
    persister = measure_persist.persister
    commands = [
        measure_persist,
        MeasureReindex(
            persister_type=persister_type,
            persister=persister,
            paths=paths,
        ),
        MeasureRetrieve(
            persister_type=persister_type,
            n=n,
            persister=persister,
            paths=paths,
        ),
        MeasureClear(
            persister_type=persister_type,
            n=n,
            persister=persister,
            paths=paths,
        ),
    ]
    for command in commands:
        delimit(delimit_tag + f"  ({command.get_name()})", "yellow1")
        rich.print(command)
        if not dry_run:
            start_time = time.time()
            data = command.measure(profile=profile)
            times.append(
                TotalTime(
                    command.get_name(),
                    time.time() - start_time if profile else data.total_time,
                )
            )
    return times


class _SuiteTotals(NamedTuple):
    persister_type: PersisterType
    profile: bool
    times: list[TotalTime]


class SuiteTotals(NamedTuple):
    without_profiling: _SuiteTotals
    with_profiling: _SuiteTotals


def persister_suite(
    persister_type: PersisterType,
    n: int = 10000,
    profile: bool = True,
    dry_run: bool = False,
) -> SuiteTotals:
    delimit(f"{persister_type.value}", "green")
    without_profiling = _persister_suite(
        persister_type=persister_type,
        n=n,
        profile=False,
        clean_measurements=True,
        dry_run=dry_run,
    )
    if profile:
        with_profiling = _persister_suite(
            persister_type=persister_type,
            n=n,
            profile=True,
            clean_measurements=False,
            dry_run=dry_run,
        )
    else:
        with_profiling = []
    return SuiteTotals(
        without_profiling=_SuiteTotals(
            persister_type=persister_type,
            profile=False,
            times=without_profiling,
        ),
        with_profiling=_SuiteTotals(
            persister_type=persister_type,
            profile=True,
            times=with_profiling,
        ),
    )


@app.command()
def suite(
    persister_type: PersisterType = PersisterType.all,
    n: int = 10000,
    profile: bool = True,
    dry_run: bool = False,
) -> None:
    """Run a suite of measurements for all persister types."""
    persister_types = (
        persister_classes.keys()
        if persister_type == PersisterType.all
        else [persister_type]
    )
    persister_names = [persister_type.value for persister_type in persister_types]
    rich.print(
        delimit_str(
            f"Running a suite of measurements for persisters {persister_names}",
            "cyan",
        ).lstrip()
    )
    totals: list[SuiteTotals] = [
        persister_suite(
            persister_type=persister_type, n=n, profile=profile, dry_run=dry_run
        )
        for persister_type in persister_types
    ]
    if not dry_run:
        if profile:
            rich.print("\nElapsed time profiling:")
            for total in totals:
                rich.print(f"  {total.with_profiling.persister_type.value}:")
                for command_total in total.with_profiling.times:
                    rich.print(
                        f"    {command_total.func_name:10s}:  {command_total.time:10.7f}"
                    )
        rich.print("\nElapsed time without profiling:")
        for total in totals:
            rich.print(f"  {total.without_profiling.persister_type}:")
            for command_total in total.without_profiling.times:
                rich.print(
                    f"    {command_total.func_name:10s}:  {command_total.time:10.7f}"
                )
    rich.print("\n")


@app.callback()
def _main() -> None: ...


if __name__ == "__main__":
    app()



"""This packages provides infrastructure for running a proactor on top of asyncio with support multiple MQTT clients
and and sub-objects which support their own threads for synchronous operations.

This packages is not GridWorks-aware (except that it links actors with multiple mqtt clients). This separation between
communication / action infrastructure and GridWorks semantics is intended to allow the latter to be more focussed.
"""

from gwproactor.actors import Actor, SyncThreadActor, SyncThreadT, WebEventListener
from gwproactor.actors.actor import PrimeActor
from gwproactor.app import App
from gwproactor.callbacks import ProactorCallbackFunctions, ProactorCallbackInterface
from gwproactor.codecs import CodecFactory, ProactorCodec
from gwproactor.config import AppSettings, ProactorSettings
from gwproactor.config.links import CodecSettings, LinkSettings
from gwproactor.config.proactor_config import ProactorConfig, ProactorName
from gwproactor.external_watchdog import ExternalWatchdogCommandBuilder
from gwproactor.links.mqtt import QOS, MQTTClients, MQTTClientWrapper, Subscription
from gwproactor.logger import ProactorLogger
from gwproactor.logging_setup import format_exceptions, setup_logging
from gwproactor.proactor_implementation import Proactor
from gwproactor.proactor_interface import (
    INVALID_IO_TASK_HANDLE,
    ActorInterface,
    AppInterface,
    Communicator,
    CommunicatorInterface,
    MonitoredName,
    Runnable,
)
from gwproactor.problems import Problems
from gwproactor.sync_thread import (
    AsyncQueueWriter,
    SyncAsyncInteractionThread,
    SyncAsyncQueueWriter,
    responsive_sleep,
)

__all__ = [
    "INVALID_IO_TASK_HANDLE",
    "QOS",
    "Actor",
    "ActorInterface",
    "App",
    "AppSettings",
    "AsyncQueueWriter",
    "CodecFactory",
    "CodecSettings",
    "Communicator",
    "CommunicatorInterface",
    "ExternalWatchdogCommandBuilder",
    "LinkSettings",
    "MQTTClientWrapper",
    "MQTTClients",
    "MonitoredName",
    "PrimeActor",
    "Proactor",
    "ProactorCallbackFunctions",
    "ProactorCallbackInterface",
    "ProactorCodec",
    "ProactorLogger",
    "ProactorConfig",
    "ProactorName",
    "ProactorSettings",
    "Problems",
    "Runnable",
    "AppInterface",
    "Subscription",
    "SyncAsyncInteractionThread",
    "SyncAsyncQueueWriter",
    "SyncThreadActor",
    "SyncThreadT",
    "WebEventListener",
    "format_exceptions",
    "responsive_sleep",
    "setup_logging",
]
